# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2020-2020  Ke Zhang <kzhang@uri.edu>

from __future__ import unicode_literals

import time
import logging

from ..command import (Command, MetabolicMixin, LoopRemovalMixin,
                       ObjectiveMixin, SolverCommandMixin,
                       ParallelTaskMixin, convert_to_unicode)
from .. import fluxanalysis

from six.moves import range

logger = logging.getLogger(__name__)


class RobustnessCommand(MetabolicMixin, LoopRemovalMixin, ObjectiveMixin,
                        SolverCommandMixin, ParallelTaskMixin, Command):
    """Run robustness analysis on the model.

    Given a reaction to maximize and a reaction to vary,
    the robustness analysis will run FBA while fixing the
    reaction to vary at each iteration. The reaction will
    be fixed at the specified number of steps between the
    minimum and maximum flux value specified in the model.
    """

    _supported_loop_removal = ['none', 'tfba', 'l1min']

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--steps', metavar='N', type=int, default=10,
            help='Number of flux value steps for varying reaction')
        parser.add_argument(
            '--minimum', metavar='V', type=float,
            help='Minumum flux value of varying reaction')
        parser.add_argument(
            '--maximum', metavar='V', type=float,
            help='Maximum flux value of varying reaction')
        parser.add_argument(
            '--all-reaction-fluxes',
            help='Print reaction flux for all model reactions',
            action='store_true')
        parser.add_argument('varying', type=convert_to_unicode,
                            help='Reaction to vary')
        parser.add_argument(
            '--fva', action='store_true',
            help='Run FVA and print flux range instead of flux value')
        super(RobustnessCommand, cls).init_parser(parser)

    def run(self):
        """Run robustness command."""

        reaction = self._get_objective()
        if not self._mm.has_reaction(reaction):
            self.fail('Specified biomass reaction is not in model: {}'.format(
                reaction))

        varying_reaction = self._args.varying
        if not self._mm.has_reaction(varying_reaction):
            self.fail('Specified varying reaction is not in model: {}'.format(
                varying_reaction))

        steps = self._args.steps
        if steps <= 0:
            self.argument_error('Invalid number of steps: {}\n'.format(steps))

        loop_removal = self._get_loop_removal_option()
        solver = self._get_solver()
        if loop_removal != 'none':
            if self._args.fva is not True:
                if loop_removal == 'tfba':
                    solver = self._get_solver(integer=True)
                else:
                    solver = self._get_solver()
            else:
                logger.warning(
                    'The loop removal method {} is not possible with --fva '
                    'option in this command'.format(loop_removal))
                quit()

        p = fluxanalysis.FluxBalanceProblem(self._mm, solver)
        if loop_removal == 'tfba':
            p.add_thermodynamic()

        try:
            p.check_constraints()
        except fluxanalysis.FluxBalanceError as e:
            self.report_flux_balance_error(e)

        # Determine minimum and maximum flux for varying reaction
        if self._args.maximum is None:
            p.maximize(varying_reaction)
            flux_max = p.get_flux(varying_reaction)
        else:
            flux_max = self._args.maximum

        if self._args.minimum is None:
            p.maximize({varying_reaction: -1})
            flux_min = p.get_flux(varying_reaction)
        else:
            flux_min = self._args.minimum

        if flux_min > flux_max:
            self.argument_error('Invalid flux range: {}, {}\n'.format(
                flux_min, flux_max))

        logger.info('Varying {} in {} steps between {} and {}'.format(
            varying_reaction, steps, flux_min, flux_max))

        start_time = time.time()

        handler_args = (
            self._mm, solver, loop_removal, self._args.all_reaction_fluxes)

        def iter_tasks():
            for i in range(steps):
                fixed_flux = flux_min + i*(flux_max - flux_min)/float(steps-1)
                constraint = varying_reaction, fixed_flux
                yield constraint, reaction

        # Run FBA on model at different fixed flux values
        if self._args.fva is not True:
            executor = self._create_executor(
                RobustnessTaskHandler, handler_args, cpus_per_worker=2)
            with executor:
                for task, result in executor.imap_unordered(iter_tasks(), 16):
                    (varying_reaction, fixed_flux), _ = task
                    if result is None:
                        logger.warning('No solution found for {} at {}'.format(
                            varying_reaction, fixed_flux))
                    elif self._args.all_reaction_fluxes:
                        for other_reaction in self._mm.reactions:
                            print('{}\t{}\t{}'.format(
                                other_reaction, fixed_flux,
                                result[other_reaction]))
                    else:
                        print('{}\t{}'.format(fixed_flux, result))

            executor.join()

            logger.info('Solving took {:.2f} seconds'.format(
                time.time() - start_time))

        # Run FVA on model at different fixed flux values
        else:
            executor_fva = self._create_executor(
                RobustnessTaskHandlerFva, handler_args, cpus_per_worker=2)
            with executor_fva:
                for task, result in executor_fva.imap_unordered(
                        iter_tasks(), 16):
                    (varying_reaction, fixed_flux), _ = task
                    if result is None:
                        logger.warning('No solution found for {} at {}'.format(
                            varying_reaction, fixed_flux))
                    elif self._args.all_reaction_fluxes:
                        for other_reaction in self._mm.reactions:
                            print('{}\t{}\t{}\t{}'.format(
                                other_reaction, fixed_flux,
                                result[other_reaction][0],
                                result[other_reaction][1]))
                    else:
                        print('{}\t{}\t{}'.format(
                            fixed_flux, result[reaction][0],
                            result[reaction][1]))

            executor_fva.join()

            logger.info('Solving took {:.2f} seconds'.format(
                time.time() - start_time))


class RobustnessTaskHandler(object):
    def __init__(self, model, solver, loop_removal, all_reactions):
        self._problem = fluxanalysis.FluxBalanceProblem(model, solver)

        if loop_removal == 'none':
            self._run_fba = self._problem.maximize
        elif loop_removal == 'l1min':
            self._run_fba = self._problem.max_min_l1
        elif loop_removal == 'tfba':
            self._problem.add_thermodynamic()
            self._run_fba = self._problem.maximize

        self._reactions = list(model.reactions) if all_reactions else None

    def handle_task(self, constraint, reaction):
        varying_reaction, fixed_flux = constraint
        flux_var = self._problem.get_flux_var(varying_reaction)
        c, = self._problem.prob.add_linear_constraints(flux_var == fixed_flux)

        try:
            self._run_fba(reaction)
            if self._reactions is not None:
                return {reaction: self._problem.get_flux(reaction)
                        for reaction in self._reactions}
            else:
                return self._problem.get_flux(reaction)
        except fluxanalysis.FluxBalanceError:
            return None
        finally:
            c.delete()


class RobustnessTaskHandlerFva(object):
    def __init__(self, model, solver, loop_removal, all_reactions):
        self._problem = fluxanalysis.FluxBalanceProblem(model, solver)

        if loop_removal == 'none':
            self._run_fba = self._problem.maximize
        # elif loop_removal == 'l1min':
        #     self._run_fba = self._problem.max_min_l1
        elif loop_removal == 'tfba':
            self._problem.add_thermodynamic()
            self._run_fba = self._problem.maximize

        self._reactions = list(model.reactions) if all_reactions else None

    def handle_task(self, constraint, reaction):
        d = None
        varying_reaction, fixed_flux = constraint
        flux_var = self._problem.get_flux_var(varying_reaction)
        c, = self._problem.prob.add_linear_constraints(flux_var == fixed_flux)
        try:
            self._problem.maximize(reaction)
            obj_flux = self._problem.get_flux(reaction)
            d, = self._problem.prob.add_linear_constraints(
                self._problem.get_flux_var(reaction) == obj_flux)
            if self._reactions is not None:
                all_rxn_fluxes = {}
                for rxn in self._reactions:
                    all_rxn_fluxes[rxn] = (
                        self._problem.flux_bound(rxn, -1),
                        self._problem.flux_bound(rxn, 1))
                return all_rxn_fluxes
            else:
                return {reaction: (self._problem.flux_bound(reaction, -1),
                        self._problem.flux_bound(reaction, 1))}
        except fluxanalysis.FluxBalanceError:
            return None
        finally:
            c.delete()
            if d is not None:
                d.delete()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2016-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2016  Chao liu <lcddzyx@gmail.com>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

import unittest

from psamm.metabolicmodel import MetabolicModel
from psamm.database import DictDatabase
from psamm import fastgapfill
from psamm.lpsolver import generic
from psamm.datasource.reaction import parse_reaction


class TestFastGapFill(unittest.TestCase):
    def setUp(self):
        self._database = DictDatabase()
        self._database.set_reaction('rxn_1', parse_reaction('|A| <=>'))
        self._database.set_reaction('rxn_2', parse_reaction('|A| <=> |B|'))
        self._database.set_reaction('rxn_3', parse_reaction('|C| <=> |B|'))
        self._database.set_reaction('rxn_4', parse_reaction('|C| <=>'))
        self._mm = MetabolicModel.load_model(
            self._database, self._database.reactions)

        try:
            self._solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_fastgapfill(self):
        core = {'rxn_2', 'rxn_3'}
        induced = fastgapfill.fastgapfill(
            self._mm,
            core,
            epsilon=0.001,
            solver=self._solver)

        self.assertEqual(
            set(induced),
            {'rxn_1', 'rxn_2', 'rxn_3', 'rxn_4'})


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

import os
import unittest
from decimal import Decimal
from fractions import Fraction

from psamm.lpsolver import generic, lp


class TestSolverProblem(unittest.TestCase):
    """Test the current solver chosen by generic.

    To test all solvers, run this test multiple times with the PSAMM_SOLVER
    environment variable set to the solver to test.
    """
    def setUp(self):
        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_set_decimal_variable_bounds(self):
        """Test that Decimal can be used for variable bounds."""
        prob = self.solver.create_problem()
        prob.define('x', lower=Decimal('3.5'), upper=Decimal('500.3'))

    def test_set_decimal_objective(self):
        """Test that Decimal can be used in objective."""
        prob = self.solver.create_problem()
        prob.define('x', lower=3, upper=100)
        prob.set_objective(Decimal('3.4') * prob.var('x'))

    def test_set_fraction_objective(self):
        """Test that Fraction can be used in objective."""
        prob = self.solver.create_problem()
        prob.define('x', lower=-5, upper=5)
        prob.set_objective(Fraction(1, 2) * prob.var('x'))

    def test_set_decimal_constraint(self):
        """Test that Decimal can be used in constraints."""
        prob = self.solver.create_problem()
        prob.define('x', lower=0, upper=5)
        prob.add_linear_constraints(
            Decimal('5.6') * prob.var('x') >= Decimal('8.9'))

    def test_result_evaluate_decimal_expression(self):
        """Test that Decimal in expression can be used to evaluate result."""
        prob = self.solver.create_problem()
        prob.define('x', lower=0, upper=Fraction(5, 2))
        prob.add_linear_constraints(prob.var('x') >= 2)
        prob.set_objective(prob.var('x'))
        result = prob.solve(lp.ObjectiveSense.Maximize)
        value = result.get_value(Decimal('3.4') * prob.var('x'))
        self.assertAlmostEqual(value, Fraction(17, 2))

    def test_redefine_variable(self):
        """Test that redefining a variable fails."""
        prob = self.solver.create_problem()
        prob.define('x', lower=3, upper=100)
        with self.assertRaises(ValueError):
            prob.define('x', lower=3, upper=100)

    def test_has_variable(self):
        """Test whether has_variable() works."""
        prob = self.solver.create_problem()
        self.assertFalse(prob.has_variable('x'))
        prob.define('x', lower=-400)
        self.assertTrue(prob.has_variable('x'))


class TestListSolversCommand(unittest.TestCase):
    def test_list_lpsolvers(self):
        # This test case is for travis and will not pass on a local computer
        if os.environ.get('PSAMM_SOLVER') in ('nosolver', None):
            with self.assertRaises(SystemExit):
                generic.list_solvers([])
        else:
            generic.list_solvers([])


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2019-2020       Jing Wang <jingwang89@uri.edu>

"""Importer for the COBRApy MAT format."""

from psamm.importer import Importer as BaseImporter, ModelLoadError
from psamm.datasource.entry import (DictCompoundEntry as CompoundEntry,
                                    DictReactionEntry as ReactionEntry,
                                    DictCompartmentEntry as CompartmentEntry)
from psamm.datasource import native
from psamm.reaction import Reaction, Compound, Direction
import re
import numpy as np
from scipy.io import loadmat
import decimal
import logging
import glob
import os


logger = logging.getLogger(__name__)


def _float_parser(num_str):
    num = float(num_str)
    if num.is_integer():
        return int(num)
    else:
        return decimal.Decimal(num_str)


class Importer(BaseImporter):
    """Read metabolic model from COBRA MATLAB mat format."""

    name = 'matlab'
    title = 'COBRA MATLAB mat'
    generic = True

    def help(self):
        """Print help text for importer."""
        print('Source must contain the model definition in COBRA MATLAB mat'
              ' format.\n'
              'Expected files in source directory:\n'
              '- *.mat')

    def _resolve_source(self, source):
        """Resolve source to filepath if it is a directory."""
        if os.path.isdir(source):
            sources = glob.glob(os.path.join(source, '*.mat'))
            if len(sources) == 0:
                raise ModelLoadError('No .mat file found in source directory')
            elif len(sources) > 1:
                raise ModelLoadError(
                    'More than one .mat file found in source directory')
            return sources[0]
        return source

    def _import(self, file, name):
        model_doc = loadmat(file)
        if name is None:
            names = [k for k in model_doc.keys()
                     if k not in ['__header__', '__version__', '__globals__']]
            modelnames = list()
            for i in names:
                if (isinstance(model_doc[i], np.ndarray) and
                        len(model_doc[i].dtype) > 1):
                    modelnames.append(i)
            if len(modelnames) == 1:
                name = str(modelnames[0])
                model_doc = model_doc[modelnames[0]]
            elif len(modelnames) > 1:
                raise ModelLoadError(
                    ('More than one model exist, '
                     'please choose from: %s '
                     'using --model-name' % modelnames))
            else:
                raise ModelLoadError('Incorrect format')
        else:
            if name not in model_doc.keys():
                raise ModelLoadError('Incorrect model name specified')
            else:
                model_doc = model_doc[name]
                if not (isinstance(model_doc, np.ndarray) and
                        len(model_doc.dtype) > 1):
                    raise ModelLoadError('Incorrect format')
        model = native.NativeModel()
        model.name = name
        self._compartments_from_compound = set()
        self._origid_compound = dict()
        model.compounds.update(self._read_compounds(model_doc))
        model.compartments.update(self._compartments_from_compound)

        self._boundary_compartment = None
        # Add model level compartment information
        if all(var in model_doc.dtype.names for var in ['comps', 'compNames']):
            model.compartments.update(self._read_compartments(model_doc))
            for comp in model.compartments:
                if comp.name.lower() == 'boundary':  # has boundary compartment
                    self._boundary_compartment = comp.id
            model.compartments.discard(self._boundary_compartment)

        for reaction, lower, upper in self._read_reactions(model_doc):
            model.reactions.add_entry(reaction)
            if lower is not None and upper is not None:
                model.limits[reaction.id] = reaction.id, lower, upper

        biomass_reaction = None
        objective_reactions = set()
        if 'c' in model_doc.dtype.names:
            for i in range(len(model_doc['c'][0, 0])):
                if model_doc['c'][0, 0][i][0] != 0:
                    objective_reactions.add(
                        str(model_doc['rxns'][0, 0][i][0][0]))
            if len(objective_reactions) == 1:
                biomass_reaction = next(iter(objective_reactions))
                logger.info('Detected biomass reaction: {}'.format(
                    biomass_reaction))
            elif len(objective_reactions) > 1:
                logger.warning(
                    'Multiple reactions are used as the'
                    ' biomass reaction: {}'.format(objective_reactions))
        else:
            logger.warning('No objective reaction')

        model.biomass_reaction = biomass_reaction

        return model

    def _read_compartments(self, doc):
        for i in range(len(doc['comps'][0, 0])):
            compartment = str(doc['comps'][0, 0][i][0][0])
            name = str(doc['compNames'][0, 0][i][0][0])
            yield CompartmentEntry(dict(id=compartment, name=name))

    def _read_compounds(self, doc):
        for i in range(len(doc['mets'][0, 0])):
            properties = dict()
            id = doc['mets'][0, 0][i][0][0]
            match = re.search(r'\[[0-9a-zA-Z]+\]$', id)  # e.g. h2o[c]
            uniqid = id
            compartment = None
            if match is not None:  # compartment information in id
                compartment = match.group().strip('[]')
                self._compartments_from_compound.add(
                    CompartmentEntry(dict(id=compartment, name='')))
                uniqid = re.sub(r'\[[0-9a-zA-Z]+\]$', '', id)
            else:
                match = re.search(r'_[0-9a-zA-Z]+$', id)  # e.g. h2o_c
                if match is not None:  # compartment information in id
                    compartment = match.group().strip('_')
                    self._compartments_from_compound.add(
                        CompartmentEntry(
                            dict(id=compartment, name='')))
                    uniqid = re.sub(r'_[0-9a-zA-Z]+$', '', id)
                else:
                    match = re.search(r'\([0-9a-zA-Z]+\)$', id)  # e.g. h2o(c)
                    if match is not None:  # compartment information in id
                        compartment = match.group().strip('()')
                        self._compartments_from_compound.add(
                            CompartmentEntry(dict(id=compartment, name='')))
                        uniqid = re.sub(r'\([0-9a-zA-Z]+\)$', '', id)
                    else:
                        match = re.search(r'[0-9][a-z]$', id)  # e.g. cpd100c
                        if match is not None:
                            compartment = str(match.group()[1])
                            self._compartments_from_compound.add(
                                CompartmentEntry(
                                    dict(id=compartment, name='')))
                            uniqid = id[:len(id) - 1]
            properties['id'] = uniqid
            self._origid_compound[id] = Compound(uniqid, compartment)
            for dtype in doc.dtype.names:
                if dtype.startswith('met') and dtype != 'mets':
                    # Compound related dtypes
                    if dtype == 'metFormulas':
                        if len(doc['metFormulas'][0, 0][i][0]) > 0:
                            properties[
                                'formula'] = doc['metFormulas'][0, 0][i][0][0]
                    elif dtype == 'metCharges':
                        properties['charge'] = _float_parser(
                            doc['metCharges'][0, 0][i][0])
                    elif dtype == 'metNames':
                        properties['name'] = str(
                            doc['metNames'][0, 0][i][0][0])
                    else:  # other dtypes
                        value = doc[dtype][0, 0][i][0]
                        if isinstance(value, np.ndarray):
                            if len(value) == 1:
                                value = value[0]
                            elif len(value) == 0:
                                continue
                        dtype = dtype.lstrip('met')
                        properties[dtype] = str(value)

            entry = CompoundEntry(properties)
            if 'formula' in entry.properties:
                formula = self._try_parse_formula(
                    entry.id, entry.properties['formula'])
                if formula is not None:
                    entry.properties['formula'] = formula
            yield entry

    def _read_reactions(self, doc):
        for i in range(len(doc['rxns'][0, 0])):
            properties = dict()
            properties['id'] = str(doc['rxns'][0, 0][i][0][0])

            for dtype in doc.dtype.names:
                if dtype.startswith('rxn') and dtype != 'rxns':
                    # reaction related dtypes
                    if dtype == 'rxnNames':
                        if len(doc['rxnNames'][0, 0][i][0]) > 0:
                            properties['name'] = str(
                                doc['rxnNames'][0, 0][i][0][0])
                    else:  # other dtypes
                        value = doc[dtype][0, 0][i][0]
                        if isinstance(value, np.ndarray):
                            if len(value) == 1:
                                value = value[0]
                            elif len(value) == 0:
                                continue
                        dtype = dtype.lstrip('rxn')
                        properties[dtype] = str(value)

            lower = float(doc['lb'][0, 0][i][0])
            upper = float(doc['ub'][0, 0][i][0])
            if lower == 0 and upper != 0:
                direction = Direction.Forward
            elif lower != 0 and upper == 0:
                direction = Direction.Reverse
            else:
                direction = Direction.Both

            # parse equation
            compounds = list()
            try:
                # incase it's a sparse matrix
                rows = doc['S'][0, 0][:, i].indices
            except AttributeError:
                # incase it's an ndarray
                rows = np.flatnonzero(doc['S'][0, 0][:, i])
            for j in rows:
                cpd = self._origid_compound.get(
                    doc['mets'][0, 0][j][0][0]
                )
                if cpd.compartment == self._boundary_compartment:
                    # skip compound that is out of boundary
                    continue
                compounds.append(
                    (
                        cpd,  # compound id
                        float(doc['S'][0, 0][j, i])  # stochiometry
                    )
                )
            properties['equation'] = (Reaction(direction, compounds))

            # parse genes
            if 'grRules' in doc.dtype.names:
                if len(doc['grRules'][0, 0][i][0]) > 0:
                    genes = doc['grRules'][0, 0][i][0][0]
                    if isinstance(genes, np.unicode_):
                        properties['genes'] = self._try_parse_gene_association(
                            properties['id'], genes
                        )

            # parse subsystem
            if 'subSystems' in doc.dtype.names:
                if doc['subSystems'][0, 0][i][0].size > 0:
                    # the subSystem has either 2 levels or 4 levels
                    # check which level to obtain
                    if isinstance(doc['subSystems'][0, 0][i][0][0], str):
                        properties['subsystem'] = str(
                            doc['subSystems'][0, 0][i][0][0])
                    elif len(doc['subSystems'][0, 0][i][0][0][0]) > 0:
                        properties['subsystem'] = str(
                            doc['subSystems'][0, 0][i][0][0][0][0])

            entry = ReactionEntry(properties)
            yield entry, lower, upper

    def import_model(self, source, model_name=None):
        """Import and return model instance."""
        if not hasattr(source, 'read'):  # Not a File-like object
            with open(self._resolve_source(source), 'rb') as f:
                return self._import(f, model_name)
        else:
            return self._import(source, model_name)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2020-2020  Elysha Sameth <esameth1@my.uri.edu>

"""Definitions related to reaction equations and parsing of such equations."""

from __future__ import unicode_literals

import re
import functools
import enum
import numbers
from collections import Counter

import six
from six import text_type, iteritems


@six.python_2_unicode_compatible
@functools.total_ordering
class Compound(object):
    """Represents a compound in a reaction equation

    A compound is a named entity in the reaction equations representing a
    chemical compound. A compound can represent a generalized chemical entity
    (e.g. polyphosphate) and the arguments can be used to instantiate a
    specific chemical entity (e.g. polyphosphate(3)) by passing a number as an
    argument or a partially specified entity by passing an expression (e.g.
    polyphosphate(n)).
    """

    def __init__(self, name, compartment=None, arguments=()):
        self._name = name
        self._compartment = None
        if compartment is not None:
            self._compartment = compartment
        self._arguments = tuple(arguments)

    @property
    def name(self):
        """Name of compound"""
        return self._name

    @property
    def compartment(self):
        """Compartment of compound"""
        return self._compartment

    @property
    def arguments(self):
        """Expression argument for generalized compounds"""
        return self._arguments

    def translate(self, func):
        """Translate compound name using given function

        >>> Compound('Pb').translate(lambda x: x.lower())
        Compound('pb')
        """
        return self.__class__(
            func(self._name), self._compartment, self._arguments)

    def in_compartment(self, compartment):
        """Return an instance of this compound in the specified compartment

        >>> Compound('H+').in_compartment('e')
        Compound('H+', 'e')
        """
        return self.__class__(self._name, compartment, self._arguments)

    def __eq__(self, other):
        return (isinstance(other, Compound) and
                self._name == other._name and
                self._compartment == other._compartment and
                self._arguments == other._arguments)

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        if isinstance(other, Compound):
            def tuple_repr(c):
                comp = c._compartment if c._compartment is not None else ''
                return (c._name, comp, c._arguments)
            return tuple_repr(self) < tuple_repr(other)
        return NotImplemented

    def __hash__(self):
        return (hash('Compound') ^ hash(self._name) ^
                hash(self._compartment) ^ hash(self._arguments))

    def __str__(self):
        """String representation of compound

        >>> str(Compound('Phosphate'))
        'Phosphate'
        >>> str(Compound('Phosphate', 'e'))
        'Phosphate[e]'
        >>> str(Compound('Polyphosphate', None, [Expression('n')]))
        'Polyphosphate(n)'
        >>> str(Compound('Polyphosphate', 'p', [Expression('n')]))
        'Polyphosphate(n)[p]'
        """
        s = self._name
        if len(self._arguments) > 0:
            s += '({})'.format(
                ', '.join(text_type(a) for a in self._arguments))
        if self._compartment is not None:
            s += '[{}]'.format(self._compartment)
        return s

    def __repr__(self):
        def str_repr(*args):
            return str('Compound({})').format(', '.join(repr(a) for a in args))

        if len(self._arguments) == 0:
            if self._compartment is None:
                return str_repr(self._name)
            return str_repr(self._name, self._compartment)
        return str_repr(self._name, self._compartment, self._arguments)


class Direction(enum.Enum):
    """Directionality of reaction equation."""
    Forward = False, True
    Reverse = True, False
    Both = True, True

    Right = False, True
    Left = True, False
    Bidir = True, True

    @property
    def forward(self):
        """Whether this direction includes forward direction."""
        return self.value[1]

    @property
    def reverse(self):
        """Whether this direction includes reverse direction."""
        return self.value[0]

    def flipped(self):
        """Return the flipped version of this direction."""
        forward, reverse = self.value
        return self.__class__((reverse, forward))

    @property
    def symbol(self):
        """Return string symbol for direction."""
        if self == Direction.Forward:
            return '=>'
        elif self == Direction.Reverse:
            return '<='
        else:
            return '<=>'


@six.python_2_unicode_compatible
class Reaction(object):
    """Reaction equation representation.

    Each compound is associated with a stoichiometric value and the reaction
    has a :class:`.Direction`. The reaction is created in one of the three
    following ways.

    It can be created from a direction and two iterables of compound,
    value pairs representing the left-hand side and the right-hand side of
    the reaction:

    >>> r = Reaction(Direction.Both, [(Compound('A'), 1), (Compound('B', 2))],
                                     [(Compound('C'), 1)])
    >>> str(r)
    '|A| + (2) |B| <=> |C|'

    It can also be created from a single dict or iterable of compound, value
    pairs where the left-hand side compounds have negative values and the
    right-hand side compounds have positive values:

    >>> r = Reaction(Direction.Forward, {
            Compound('A'): -1,
            Compound('B'): -2,
            Compound('C'): 1
    })
    >>> str(r)
    '|A| + (2) |B| <=> |C|'

    Lastly, the reaction can be created from an existing reaction object,
    creating a copy of that reaction.

    >>> r = Reaction(Direction.Forward, {Compound('A'): -1, Compound('B'): 1})
    >>> r2 = Reaction(r)
    >>> str(r2)
    '|A| => |B|'

    Reactions can be added to produce combined reactions.

    >>> r = Reaction(Direction.Forward, {Compound('A'): -1, Compound('B'): 1})
    >>> s = Reaction(Direction.Forward, {Compound('B'): -1, Compound('C'): 1})
    >>> str(r + s)
    '|A| => |C|'

    Reactions can also be multiplied by a number to produce a new reaction
    with scaled stoichiometry.

    >>> r = Reaction(Direction.Forward, {Compound('A'): -1, Compound('B'): 2})
    >>> str(2 * r)
    '(2) |A| => (4) |B|'

    Multiplying with a negative value will also flip the reaction, and as a
    special case, negating a reaction will simply flip it.

    >>> r = Reaction(Direction.Forward, {Compound('A'): -1, Compound('B'): 2})
    >>> str(r)
    '|A| => (2) |B|'
    >>> str(-r)
    '(2) |B| <= |A|'
    """

    def __init__(self, *args):
        if len(args) == 1:
            # Initialize from Reaction object
            if not isinstance(args[0], self.__class__):
                raise TypeError('Single argument must be of type {}'.format(
                    self.__class__.__name__))

            self._direction = args[0].direction
            self._left = tuple(args[0].left)
            self._right = tuple(args[0].right)
        elif len(args) == 2:
            # Initialize from direction and dict or single iterable
            if not isinstance(args[0], Direction):
                raise TypeError('First argument must be a Direction')
            self._direction = args[0]

            values = args[1]
            if isinstance(values, dict):
                values = iteritems(values)

            left, right = [], []
            for compound, value in values:
                if not isinstance(value, numbers.Number):
                    raise TypeError('Values must be numeric')
                if value < 0:
                    left.append((compound, -value))
                elif value > 0:
                    right.append((compound, value))

            self._left = tuple(left)
            self._right = tuple(right)
        elif len(args) == 3:
            # Initialize from direction and two iterables
            if not isinstance(args[0], Direction):
                raise TypeError('First argument must be a Direction')
            self._direction = args[0]

            left, right = [], []
            for compound, value in args[1]:
                if isinstance(value, numbers.Number) and value < 0:
                    raise ValueError('Value must not be negative')
                elif value != 0:
                    left.append((compound, value))

            for compound, value in args[2]:
                if isinstance(value, numbers.Number) and value < 0:
                    raise ValueError('Value must not be negative')
                elif value != 0:
                    right.append((compound, value))

            self._left = tuple(left)
            self._right = tuple(right)
        else:
            raise TypeError('Too many arguments (one, two or three'
                            ' arguments required)')

    @property
    def direction(self):
        """Direction of reaction equation"""
        return self._direction

    @property
    def left(self):
        """Compounds on the left-hand side of the reaction equation."""
        return self._left

    @property
    def right(self):
        """Compounds on the right-hand side of the reaction equation."""
        return self._right

    @property
    def compounds(self):
        """Sequence of compounds on both sides of the reaction equation

        The sign of the stoichiometric values reflect whether the compound is
        on the left-hand side (negative) or the right-hand side (positive).
        """
        return tuple((c, -v) for c, v in self._left) + self._right

    def normalized(self):
        """Return normalized reaction

        The normalized reaction will be bidirectional or a forward reaction
        (i.e. reverse reactions are flipped).
        """

        if self._direction == Direction.Reverse:
            return self.__class__(
                self._direction.flipped(), self._right, self._left)

        return self

    def translated_compounds(self, translate):
        """Return reaction where compound names have been translated.

        For each compound the translate function is called with the compound
        name and the returned value is used as the new compound name. A new
        reaction is returned with the substituted compound names.
        """
        compounds = ((compound.translate(translate), value)
                     for compound, value in self.compounds)
        return Reaction(self._direction, compounds)

    def __str__(self):
        # Use the same format as ModelSEED
        def format_compound(compound, count):
            """Format compound"""
            cpdspec = text_type(compound)

            if re.search(r'\s', cpdspec):
                cpdspec = '|{}|'.format(cpdspec)
            if count != 1:
                return '({}) {}'.format(count, cpdspec)
            return cpdspec

        def format_compound_list(cmpds):
            """Format compound list"""
            return ' + '.join(format_compound(compound, count)
                              for compound, count in cmpds)

        s = text_type(self._direction.symbol)
        if len(self._left) > 0:
            s = format_compound_list(self._left) + ' ' + s
        if len(self._right) > 0:
            s += ' ' + format_compound_list(self._right)
        return s

    def __repr__(self):
        return str('Reaction({}, {}, {})').format(
            repr(self._direction), repr(self._left), repr(self._right))

    def __eq__(self, other):
        """Indicate equality of self and other"""
        return (isinstance(other, self.__class__) and
                self._direction == other._direction and
                self._left == other._left and
                self._right == other._right)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return (hash('Reaction') ^ hash(self._direction) ^
                hash(self._left) ^ hash(self._right))

    def __add__(self, other):
        if isinstance(other, Reaction):
            reverse = self.direction.reverse and other.direction.reverse
            forward = self.direction.forward and other.direction.forward
            if not reverse and not forward:
                raise ValueError('Reactions have incompatible directions')

            direction = Direction((reverse, forward))
            values = Counter(dict(self.compounds))
            values.update(dict(other.compounds))
            return Reaction(direction, values)

        return NotImplemented

    def __sub__(self, other):
        return self + -other

    def __neg__(self):
        return -1 * self

    def __mul__(self, other):
        if isinstance(other, numbers.Number):
            direction = (
                self.direction if other > 0 else self.direction.flipped())
            return self.__class__(
                direction, ((c, other * v) for c, v in self.compounds))

        return NotImplemented

    def __rmul__(self, other):
        return self * other



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Identify blocked metabolites and possible reconstructions.

This implements a variant of the algorithms described in [Kumar07]_.
"""

import logging

from six import iteritems, raise_from

from .lpsolver import lp

logger = logging.getLogger(__name__)


class GapFillError(Exception):
    """Indicates an error while running GapFind/GapFill"""


def _find_integer_tolerance(epsilon, v_max, min_tol):
    """Find appropriate integer tolerance for gap-filling problems."""
    int_tol = min(epsilon / (10 * v_max), 0.1)
    min_tol = max(1e-10, min_tol)
    if int_tol < min_tol:
        eps_lower = min_tol * 10 * v_max
        logger.warning(
            'When the maximum flux is {}, it is recommended that'
            ' epsilon > {} to avoid numerical issues with this'
            ' solver. Results may be incorrect with'
            ' the current settings!'.format(v_max, eps_lower))
        return min_tol

    return int_tol


def gapfind(model, solver, epsilon=0.001, v_max=1000, implicit_sinks=True):
    """Identify compounds in the model that cannot be produced.

    Yields all compounds that cannot be produced. This method
    assumes implicit sinks for all compounds in the model so
    the only factor that influences whether a compound can be
    produced is the presence of the compounds needed to produce it.

    Epsilon indicates the threshold amount of reaction flux for the products
    to be considered non-blocked. V_max indicates the maximum flux.

    This method is implemented as a MILP-program. Therefore it may
    not be efficient for larger models.

    Args:
        model: :class:`MetabolicModel` containing core reactions and reactions
            that can be added for gap-filling.
        solver: MILP solver instance.
        epsilon: Threshold amount of a compound produced for it to not be
            considered blocked.
        v_max: Maximum flux.
        implicit_sinks: Whether implicit sinks for all compounds are included
            when gap-filling (traditional GapFill uses implicit sinks).
    """
    prob = solver.create_problem()

    # Set integrality tolerance such that w constraints are correct
    min_tol = prob.integrality_tolerance.min
    int_tol = _find_integer_tolerance(epsilon, v_max, min_tol)
    if int_tol < prob.integrality_tolerance.value:
        prob.integrality_tolerance.value = int_tol

    # Define flux variables
    v = prob.namespace()
    for reaction_id in model.reactions:
        lower, upper = model.limits[reaction_id]
        v.define([reaction_id], lower=lower, upper=upper)

    # Define constraints on production of metabolites in reaction
    w = prob.namespace(types=lp.VariableType.Binary)
    binary_cons_lhs = {compound: 0 for compound in model.compounds}
    for spec, value in iteritems(model.matrix):
        compound, reaction_id = spec
        if value != 0:
            w.define([spec])
            w_var = w(spec)

            lower, upper = (float(x) for x in model.limits[reaction_id])
            if value > 0:
                dv = v(reaction_id)
            else:
                dv = -v(reaction_id)
                lower, upper = -upper, -lower

            prob.add_linear_constraints(
                dv <= upper * w_var,
                dv >= epsilon + (lower - epsilon) * (1 - w_var))

            binary_cons_lhs[compound] += w_var

    xp = prob.namespace(model.compounds, types=lp.VariableType.Binary)
    objective = xp.sum(model.compounds)
    prob.set_objective(objective)

    for compound, lhs in iteritems(binary_cons_lhs):
        prob.add_linear_constraints(lhs >= xp(compound))

    # Define mass balance constraints
    massbalance_lhs = {compound: 0 for compound in model.compounds}
    for spec, value in iteritems(model.matrix):
        compound, reaction_id = spec
        massbalance_lhs[compound] += v(reaction_id) * value
    for compound, lhs in iteritems(massbalance_lhs):
        if implicit_sinks:
            # The constraint is merely >0 meaning that we have implicit sinks
            # for all compounds.
            prob.add_linear_constraints(lhs >= 0)
        else:
            prob.add_linear_constraints(lhs == 0)

    # Solve
    try:
        result = prob.solve(lp.ObjectiveSense.Maximize)
    except lp.SolverError as e:
        raise_from(GapFillError('Failed to solve gapfill: {}'.format(e), e))

    for compound in model.compounds:
        if result.get_value(xp(compound)) < 0.5:
            yield compound


def gapfill(
        model, core, blocked, exclude, solver, epsilon=0.001, v_max=1000,
        weights={}, implicit_sinks=True, allow_bounds_expansion=False):
    """Find a set of reactions to add such that no compounds are blocked.

    Returns two iterators: first an iterator of reactions not in
    core, that were added to resolve the model. Second, an
    iterator of reactions in core that had flux bounds expanded (i.e.
    irreversible reactions become reversible). Similarly to
    GapFind, this method assumes, by default, implicit sinks for all compounds
    in the model so the only factor that influences whether a compound
    can be produced is the presence of the compounds needed to produce
    it. This means that the resulting model will not necessarily be
    flux consistent.

    This method is implemented as a MILP-program. Therefore it may
    not be efficient for larger models.

    Args:
        model: :class:`MetabolicModel` containing core reactions and reactions
            that can be added for gap-filling.
        core: The set of core (already present) reactions in the model.
        blocked: The compounds to unblock.
        exclude: Set of reactions in core to be excluded from gap-filling (e.g.
            biomass reaction).
        solver: MILP solver instance.
        epsilon: Threshold amount of a compound produced for it to not be
            considered blocked.
        v_max: Maximum flux.
        weights: Dictionary of weights for reactions. Weight is the penalty
            score for adding the reaction (non-core reactions) or expanding the
            flux bounds (all reactions).
        implicit_sinks: Whether implicit sinks for all compounds are included
            when gap-filling (traditional GapFill uses implicit sinks).
        allow_bounds_expansion: Allow flux bounds to be expanded at the cost
            of a penalty which can be specified using weights (traditional
            GapFill does not allow this). This includes turning irreversible
            reactions reversible.
    """
    prob = solver.create_problem()

    # Set integrality tolerance such that w constraints are correct
    min_tol = prob.integrality_tolerance.min
    int_tol = _find_integer_tolerance(epsilon, v_max, min_tol)
    if int_tol < prob.integrality_tolerance.value:
        prob.integrality_tolerance.value = int_tol

    # Define flux variables
    v = prob.namespace(model.reactions, lower=-v_max, upper=v_max)

    # Add binary indicator variables
    database_reactions = set(model.reactions).difference(core, exclude)
    ym = prob.namespace(model.reactions, types=lp.VariableType.Binary)
    yd = prob.namespace(database_reactions, types=lp.VariableType.Binary)

    objective = ym.expr(
        (rxnid, weights.get(rxnid, 1)) for rxnid in model.reactions)
    objective += yd.expr(
        (rxnid, weights.get(rxnid, 1)) for rxnid in database_reactions)
    prob.set_objective(objective)

    # Add constraints on all reactions
    for reaction_id in model.reactions:
        lower, upper = (float(x) for x in model.limits[reaction_id])

        if reaction_id in exclude or not allow_bounds_expansion:
            prob.add_linear_constraints(
                upper >= v(reaction_id), v(reaction_id) >= lower)
        else:
            # Allow flux bounds to expand up to v_max with penalty
            delta_lower = min(0, -v_max - lower)
            delta_upper = max(0, v_max - upper)
            prob.add_linear_constraints(
                v(reaction_id) >= lower + ym(reaction_id) * delta_lower,
                v(reaction_id) <= upper + ym(reaction_id) * delta_upper)

    # Add constraints on database reactions
    for reaction_id in database_reactions:
        lower, upper = model.limits[reaction_id]
        prob.add_linear_constraints(
            v(reaction_id) >= yd(reaction_id) * -v_max,
            v(reaction_id) <= yd(reaction_id) * v_max)

    # Define constraints on production of blocked metabolites in reaction
    w = prob.namespace(types=lp.VariableType.Binary)
    binary_cons_lhs = {compound: 0 for compound in blocked}
    for (compound, reaction_id), value in iteritems(model.matrix):
        if reaction_id not in exclude and compound in blocked and value != 0:
            w.define([(compound, reaction_id)])
            w_var = w((compound, reaction_id))

            dv = v(reaction_id) if value > 0 else -v(reaction_id)
            prob.add_linear_constraints(
                dv <= v_max * w_var,
                dv >= epsilon + (-v_max - epsilon) * (1 - w_var))

            binary_cons_lhs[compound] += w_var

    for compound, lhs in iteritems(binary_cons_lhs):
        prob.add_linear_constraints(lhs >= 1)

    # Define mass balance constraints
    massbalance_lhs = {compound: 0 for compound in model.compounds}
    for (compound, reaction_id), value in iteritems(model.matrix):
        if reaction_id not in exclude:
            massbalance_lhs[compound] += v(reaction_id) * value
    for compound, lhs in iteritems(massbalance_lhs):
        if implicit_sinks:
            # The constraint is merely >0 meaning that we have implicit sinks
            # for all compounds.
            prob.add_linear_constraints(lhs >= 0)
        else:
            prob.add_linear_constraints(lhs == 0)

    # Solve
    try:
        prob.solve(lp.ObjectiveSense.Minimize)
    except lp.SolverError as e:
        raise_from(GapFillError('Failed to solve gapfill: {}'.format(e)), e)

    def added_iter():
        for reaction_id in database_reactions:
            if yd.value(reaction_id) > 0.5:
                yield reaction_id

    def no_bounds_iter():
        for reaction_id in model.reactions:
            if ym.value(reaction_id) > 0.5:
                yield reaction_id

    return added_iter(), no_bounds_iter()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2017-2020  Keith Dufault-Thompson <keitht547@uri.edu>

from __future__ import unicode_literals
import logging
from ..command import SolverCommandMixin, MetabolicMixin, \
    Command, ObjectiveMixin, convert_to_unicode
from ..reaction import Compound
from ..lpsolver import lp
from ..datasource.reaction import parse_reaction
from decimal import Decimal
import csv
import math
import random
import yaml
import argparse
from psamm.datasource.entry import (DictReactionEntry as ReactionEntry)
from six import iteritems
logger = logging.getLogger(__name__)


class TMFACommand(MetabolicMixin, SolverCommandMixin, ObjectiveMixin, Command):
    """Run TMFA on the model."""

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument('--config', type=argparse.FileType('r'),
                            help='Config file for TMFA settings')
        parser.add_argument('--threshold', default=None, type=Decimal,
                            help='value to fix biomass flux to during '
                                 'tmfa simulations (default = max biomass)')
        parser.add_argument('--temp', help='Temperature in Celsius',
                            default=25)
        parser.add_argument('--hamilton', action='store_true',
                            help='run model using Hamilton TMFA method')
        parser.add_argument('--err', action='store_true',
                            help='use error estimates when running TMFA')
        parser.add_argument('--phin', default='4,11',
                            help='Allowable range for internal compartment pH '
                            'format: Lower,Upper')
        parser.add_argument('--phout', default='4,11',
                            help='Allowable range for external compartment pH '
                            'format: Lower,Upper')

        subparsers = parser.add_subparsers(title='Functions', dest='which')
        parser_util = subparsers.add_parser('util', help='Utility functions')

        parser_util.add_argument('--generate-config', action='store_true')
        parser_util.add_argument('--random-addition', action='store_true',
                                 help='perform random reaction constraint '
                                      'addition in model')

        parser_sim = subparsers.add_parser('simulation',
                                           help='Simulation Functions')
        parser_sim.add_argument('--single-solution', type=str, default=None,
                                choices=['fba', 'l1min', 'random'],
                                help='get a single TMFA solution')

        super(TMFACommand, cls).init_parser(parser)

    def run(self):
        """Run TMFA command."""
        solver = self._get_solver()
        if solver._properties['name'] not in ['cplex', 'gurobi']:
            raise ImportError(
                'Unable to find an LP solver that satisfy the requirement of '
                'TMFA function (TMFA requires Cplex or Gurobi as LP solver)')

        mm = self._mm
        which_command = self._args.which

        if which_command == 'util':
            if self._args.generate_config:
                with open('./example-config.yaml', mode='w') as f:
                    f.write('deltaG: <path to deltaG input file>\n')
                    f.write('exclude: <path to excluded reactions file>\n')
                    f.write('transporters: <path to transporter parameters '
                            'file>\n')
                    f.write('concentrations: <path to set concentrations '
                            'file>\n')
                    f.write('proton-in: <id of the internal proton. '
                            'example cpd_h[c]>\n')
                    f.write('proton-out: <id of the external proton. '
                            'example cpd_h[p]>\n')
                    f.write('proton-other:\n')
                    f.write('  - <id of other proton compounds in other '
                            'compartments>\n')
                    f.write('water:\n')
                    f.write('  - <id of water compounds example cpd_h2o[c]. '
                            'list one per line\n')
                logger.info('Generated example config file '
                            'example-config.yaml')
                quit()

        if self._args.config is None:
            logger.warning('No configuration file was provided with'
                           ' --config option')
            quit()
        else:
            config_dict = yaml.safe_load(self._args.config)

        # Parse exclude file provided through command line.
        base_exclude_list = []
        if config_dict.get('exclude') is not None:
            for line in open(config_dict.get('exclude')).readlines():
                base_exclude_list.append(line.rstrip())

        # Parse set of reactions that will have their deltaG
        # constrained only by deltaG of the transport component.
        ph_difference_rxn = []
        if config_dict.get('rxn_conc_only') is not None:
            for line in open(config_dict.get('rxn_conc_only')).readlines():
                rxn = line.rstrip()
                ph_difference_rxn.append(rxn)
                ph_difference_rxn.append(u'{}_forward'.format(rxn))
                ph_difference_rxn.append(u'{}_reverse'.format(rxn))

        # Parse out the file of lumped reactions or set the lumps
        # dictionaries to be empty
        if config_dict.get('lumped_reactions') is not None:
            lumpid_to_rxn, rxn_lumpid, lump_to_rxnids, \
                lump_to_rxnids_dir = lump_parser(
                    open(config_dict.get('lumped_reactions')))
        else:
            logger.info('No lumped reactions were provided')
            lumpid_to_rxn = {}
            lump_to_rxnids = {}
            lump_to_rxnids_dir = {}

        # Add exchange reactions to the exclude list
        for reaction in mm.reactions:
            if mm.is_exchange(reaction):
                base_exclude_list.append(reaction)

        # these lines will add ph_difference list to the base exclude list.
        base_exclude_list = base_exclude_list + ph_difference_rxn

        # exclude lump_list is a list of all excluded reactions and
        # lump reactions. exclude_lump_unkown list is a list of all
        # excluded reactions, lump reactions and lumped reactions.
        exclude_lump_list = set(base_exclude_list)
        exclude_unknown_list = set(base_exclude_list)
        for lump_id, sub_rxn_list in iteritems(lump_to_rxnids):
            exclude_lump_list.add(lump_id)
            for sub_rxn in sub_rxn_list:
                exclude_unknown_list.add(sub_rxn)
        # Make list of all excluded, lumped, and unknown dgr.
        exclude_lump_unkown = exclude_unknown_list.union(exclude_lump_list)

        cpd_conc_dict = {}
        # Parse file containing set concentrations and ranges.
        if config_dict.get('concentrations') is not None:
            for row in csv.reader(
                open(config_dict.get('concentrations')),
                    delimiter=str('\t')):
                cpd, lower, upper = row
                cpd_conc_dict[convert_to_unicode(cpd)] = [lower, upper]

        # Add lump reactions to the mm_irreversible model
        for lump_id, rxn in iteritems(lumpid_to_rxn):
            reaction = parse_reaction(rxn)
            mm.database.set_reaction(lump_id, reaction)
            mm.add_reaction(lump_id)
            mm.limits[lump_id].lower = 0
            mm.limits[lump_id].upper = 0

        # make an irreversible version of the metabolic model:
        mm_irreversible, _, split_reversible, \
            reversible_lump_to_rxn_dict = mm.make_irreversible(
                gene_dict={}, exclude_list=exclude_lump_unkown,
                lumped_rxns=lump_to_rxnids_dir,
                all_reversible=self._args.hamilton)

        split_reactions = []
        for_rev_reactions = []
        for (i, j) in split_reversible:
            for_rev_reactions.append(i)
            for_rev_reactions.append(j)
            split_reactions.append(i[:-8])

        model = self._model

        for rx in [i for i in model.reactions]:
            if rx.id in split_reactions:
                f = ReactionEntry(
                    dict(id=u'{}_forward'.format(rx.id), genes=rx.genes))
                r = ReactionEntry(
                    dict(id=u'{}_reverse'.format(rx.id), genes=rx.genes))
                model.reactions.add_entry(f)
                model.reactions.add_entry(r)

        # update these lists for the new reversible lumps and constituent
        # reactions. exclude lump_list is a list of all excluded reactions
        # and lump reactions. exclude_lump_unkown list is a list of
        # all excluded reactions, lump reactions and lumped reactions.
        for lump_id, sub_rxn_list in iteritems(reversible_lump_to_rxn_dict):
            exclude_lump_list.add(lump_id)
            for sub_rxn in sub_rxn_list:
                exclude_unknown_list.add(sub_rxn)
        # Make list of all excluded, lumped, and unknown dgr.
        exclude_lump_unkown = exclude_unknown_list.union(exclude_lump_list)

        # Read DeltaG information from file.
        dgr_dict = {}
        if config_dict.get('deltaG') is not None:
            dgr_dict = parse_dgr_file(open(
                config_dict.get('deltaG')), mm_irreversible)
            print('using dgr file')
        elif config_dict.get('deltaGf') is not None:
            # Parse the deltaGf values for all metabolites
            # from a supplied file.
            dgf_dict = parse_dgf(mm_irreversible, open(
                config_dict.get('deltaGf')))
            # Calculate the deltaGr values for all reactions
            dgr_dict = calculate_dgr(
                mm_irreversible, dgf_dict, exclude_unknown_list,
                open(config_dict.get('transporters')),
                ph_difference_rxn, open(config_dict.get('scaled_compounds')))
            print('using dgf file')

        if config_dict.get('transporters') is not None:
            transport_parameters = parse_tparam_file(
                open(config_dict.get('transporters')))

        prob, v, zi, dgri, xij, cp_list = make_tmfa_problem(
            mm_irreversible, solver)

        # Parse pH settings from command line
        phin_s = self._args.phin
        phout_s = self._args.phout
        phinsplit = phin_s.split(',')
        phoutsplit = phout_s.split(',')
        phin = (float(phinsplit[0]), float(phinsplit[1]))
        phout = (float(phoutsplit[0]), float(phoutsplit[1]))
        ph_bounds = [phin, phout]

        # Run Utility based functions
        if which_command == 'util':
            if self._args.random_addition:
                testing_list = list(mm_irreversible.reactions)
                random.shuffle(testing_list)
                testing_list_tmp = []
                for rx in testing_list:
                    testing_list_iter = testing_list_tmp + [rx]
                    logger.info(u'current testing list: {}'.format(
                        testing_list_tmp))
                    logger.info(u'testing reaction: {}'.format(rx))

                    prob, v, zi, dgri, xij, cp_list = make_tmfa_problem(
                        mm_irreversible, solver)

                    prob, cpd_xij_dict = add_conc_constraints(
                        xij, prob, cpd_conc_dict, cp_list,
                        config_dict.get('water'), config_dict.get('proton-in'),
                        config_dict.get('proton-out'),
                        config_dict.get('proton-other'))

                    prob, excluded_compounds = add_reaction_constraints(
                        prob, v, zi, dgri, xij, mm_irreversible,
                        exclude_lump_list, exclude_unknown_list,
                        exclude_lump_unkown, dgr_dict,
                        reversible_lump_to_rxn_dict, split_reversible,
                        transport_parameters, testing_list_iter,
                        config_dict.get('scaled_compounds'),
                        config_dict.get('water'), config_dict.get('proton-in'),
                        config_dict.get('proton-out'),
                        config_dict.get('proton-other'), ph_bounds,
                        self._args.temp, self._args.err)
                    try:
                        biomass = get_var_bound(prob, v(self._get_objective()),
                                                lp.ObjectiveSense.Maximize)
                        logger.info(u'Current Biomass: {}'.format(biomass))
                        if biomass < self._args.threshold:
                            continue
                        else:
                            testing_list_tmp.append(rx)
                    except:
                        continue

                for rx in testing_list:
                    if rx not in testing_list_tmp:
                        print(u'{}\tBad Constraint'.format(rx))
                    else:
                        print(u'{}\tGood Constraint'.format(rx))
                quit()
        # Run simulation based functions
        elif which_command == 'simulation':
            prob, cpd_xij_dict = add_conc_constraints(
                xij, prob, cpd_conc_dict, cp_list,
                config_dict.get('water'), config_dict.get('proton-in'),
                config_dict.get('proton-out'), config_dict.get('proton-other'))
            testing_list_tmp = list(mm_irreversible.reactions)
            prob, excluded_compounds = add_reaction_constraints(
                prob, v, zi, dgri, xij, mm_irreversible, exclude_lump_list,
                exclude_unknown_list, exclude_lump_unkown, dgr_dict,
                reversible_lump_to_rxn_dict, split_reversible,
                transport_parameters, testing_list_tmp,
                config_dict.get('scaled_compounds'), config_dict.get('water'),
                config_dict.get('proton-in'), config_dict.get('proton-out'),
                config_dict.get('proton-other'), ph_bounds,
                self._args.temp, self._args.err)

            if self._args.threshold is not None:
                prob.add_linear_constraints(
                    v(self._get_objective()) == self._args.threshold)
                logger.info(u'Set biomass based on threshold to '
                            '{}'.format(self._args.threshold))
            else:
                max_biomass = get_var_bound(prob, v(self._get_objective()),
                                            lp.ObjectiveSense.Maximize)
                prob.add_linear_constraints(
                    v(self._get_objective()) == max_biomass)
                logger.info(u'Set biomass based on maxbiomass to {}'.format(
                    max_biomass))
            split_reversible_list = []
            for (f, r) in split_reversible:
                split_reversible_list.append(f)
                split_reversible_list.append(r)
            if self._args.single_solution is not None:
                print_fba(self._args.single_solution, prob,
                          self._get_objective(), v, zi, dgri, xij,
                          mm_irreversible, cp_list, excluded_compounds,
                          split_reversible_list)
            else:
                print_fva(prob, mm_irreversible, cp_list, exclude_unknown_list,
                          v, dgri, zi, xij, excluded_compounds,
                          split_reversible_list)
        quit()


def make_tmfa_problem(mm_irreversible, solver):
    """Make a flux balance problem with TMFA related variables

    The TMFA problem contains 4 classes of variables. Flux variables
    defined in the v namespace, binary reaction indicator variables
    defined in the zi namespace, Gibbs free energy variables defined
    in the dgri namespace, and concentration variables defined in the
    xij namespace. This function will add these variables with their
    default lower and upper bounds. The function will then return
    the flux problem, variable namespaces and a list of the compounds.

    Args:
        mm_irreversible: :class:`psamm.metabolicmodel.MetabolicModel`.
        solver: linear programming library to use.
    """
    prob = solver.create_problem()
    if solver._properties['name'] == 'gurobi':
        prob.integrality_tolerance.value = 1e-9
        logger.warning(
            'Gurobi supports minimum integrality tolerance of 1e-9. This may '
            'affect the results from this simulation')
    elif solver._properties['name'] == 'glpk':
        prob.integrality_tolerance.value = 1e-21
    else:
        prob.integrality_tolerance.value = 0
    v = prob.namespace(name='flux')
    zi = prob.namespace(name='zi')
    dgri = prob.namespace(name='dgri')
    xij = prob.namespace(name='xij')
    for reaction in mm_irreversible.reactions:
        lower, upper = mm_irreversible.limits[reaction]
        v.define([reaction], lower=lower, upper=upper,
                 types=lp.VariableType.Continuous)
        zi.define([reaction], lower=int(0), upper=int(1),
                  types=lp.VariableType.Binary)
        dgri.define([reaction], lower=-1000, upper=1000,
                    types=lp.VariableType.Continuous)

    massbalance_lhs = {compound: 0 for compound in mm_irreversible.compounds}
    for spec, value in iteritems(mm_irreversible.matrix):
        compound, reaction_id = spec
        massbalance_lhs[compound] += v(reaction_id) * value
    for compound, lhs in iteritems(massbalance_lhs):
        prob.add_linear_constraints(lhs == 0)

    cp_list = []
    for cp in mm_irreversible.compounds:
        cp_list.append(convert_to_unicode(str(cp)))
        xij.define([convert_to_unicode(str(cp))], lower=-50, upper=50,
                   types=lp.VariableType.Continuous)
    return prob, v, zi, dgri, xij, cp_list


def get_var_bound(prob, var, objective_sense):
    """Gets upper or lower bound of a variable in an LP problem.

    Args:
        prob: :class:`psamm.lpsolver.lp.Problem`.
        var: LP problem variable
        objective_sense: :class:`psamm.lpsolver.lp.ObjectiveSense`
    """
    prob.set_objective(var)
    result = prob.solve_unchecked(objective_sense)
    if not result.success:
        logger.error(u'Solution not optimal: {}'.format(result.status))
        quit()
    return result.get_value(var)


def print_fva(prob, mm_irreversible, cp_list, exclude_unknown_list,
              _v, _dgri, _zi, _xij, excluded_compounds, split_reversible_list):
    """Prints FVA like result from TMFA problem.

    This function will take a TMFA problem along with the associated
    variable namespaces and print out all of the upper and lower bounds
    of the variables.

    Args:
        prob: :class:`psamm.lpsolver.lp.Problem`.
        mm_irreversible: :class:`psamm.metabolicmodel.MetabolicModel`.
        cp_list: List of compounds in the metabolic model.
        exclude_unknown_list: List of reactions excluded from thermodynamic
            constraints.
        _v: variable namespace for flux variables.
        _dgri: variable namespace for gibbs free energy variables
        _zi: variables namespace for indicator variables
        _xij: variable namespace for concentrations
        excluded_compounds: list of compounds that are not constrained
        split_reversible_list: list of all split reactions

    """
    for step, reaction in enumerate(sorted(mm_irreversible.reactions)):
        # logger.info('Testing Reaction {}/{}'.format(step, len_rxns))
        min_flux = get_var_bound(prob, _v(reaction),
                                 lp.ObjectiveSense.Minimize)
        max_flux = get_var_bound(prob, _v(reaction),
                                 lp.ObjectiveSense.Maximize)
        print(u'Flux\t{}\t{}\t{}'.format(reaction, min_flux, max_flux))
        if reaction not in exclude_unknown_list:
            if reaction in split_reversible_list:
                if '_reverse' in reaction:
                    rxn_f = reaction
                    rxn_f = rxn_f.replace('_reverse', '')
                    rxn_f += '_forward'
                    max_dgr = -1*get_var_bound(prob, _dgri(rxn_f),
                                               lp.ObjectiveSense.Minimize)
                    min_dgr = -1*get_var_bound(prob, _dgri(rxn_f),
                                               lp.ObjectiveSense.Maximize)
                else:
                    min_dgr = get_var_bound(prob, _dgri(reaction),
                                            lp.ObjectiveSense.Minimize)
                    max_dgr = get_var_bound(prob, _dgri(reaction),
                                            lp.ObjectiveSense.Maximize)
            else:
                min_dgr = get_var_bound(prob, _dgri(reaction),
                                        lp.ObjectiveSense.Minimize)
                max_dgr = get_var_bound(prob, _dgri(reaction),
                                        lp.ObjectiveSense.Maximize)
            min_zi = get_var_bound(prob, _zi(reaction),
                                   lp.ObjectiveSense.Minimize)
            max_zi = get_var_bound(prob, _zi(reaction),
                                   lp.ObjectiveSense.Maximize)
            print(u'DGR\t{}\t{}\t{}'.format(reaction, min_dgr, max_dgr))
            print(u'Zi\t{}\t{}\t{}'.format(reaction, min_zi, max_zi))
        else:
            print(u'DGR\t{}\t{}\t{}'.format(reaction, 'NA', 'NA'))
            print(u'Zi\t{}\t{}\t{}'.format(reaction, 'NA', 'NA'))
    for step, cpd in enumerate(sorted(cp_list)):
        if cpd not in excluded_compounds:
            # logger.info('Testing Compound {}\{}'.format(step, len(cp_list)))
            min_cpd = get_var_bound(prob, _xij(cpd),
                                    lp.ObjectiveSense.Minimize)
            max_cpd = get_var_bound(prob, _xij(cpd),
                                    lp.ObjectiveSense.Maximize)
            print(u'CONC\t{}\t{}\t{}'.format(
                cpd, math.exp(min_cpd), math.exp(max_cpd)))


def print_fba(simulation, prob, objective, _v, _zi, _dgri, _xij, mm,
              cp_list, excluded_compounds, split_reversible_list):
    """Prints FBA like result from TMFA problem.

    This function will take a TMFA problem along with the associated
    variable namespaces and print out a single solution for the
    associated problem. The solution can be either a single FBA-like
    solution, a L1min solution, or a random solution from the
    boundary of the solution space.

    Args:
        simulation: type of solving for the problem. ['fba', 'l1min', 'random']
        prob: :class:`psamm.lpsolver.lp.Problem`.
        objective: Objective reaction ID.
        mm_irreversible: :class:`psamm.metabolicmodel.MetabolicModel`.
        cp_list: List of compounds in the metabolic model.
        exclude_unknown_list: List of reactions excluded
            from thermo constraints.
        _v: variable namespace for flux variables.
        _dgri: variable namespace for gibbs free energy variables
        _zi: variables namespace for indicator variables
        _xij: variable namespace for concentrations
        excluded_compounds: list of compounds that are not constrained
        split_reversible_list: list of all split reactions
    """
    prob.set_objective(_v(objective))
    result = prob.solve_unchecked(lp.ObjectiveSense.Maximize)
    if not result.success:
        logger.error(u'Solution not optimal: {}'.format(result.status))
        quit()
    if simulation == 'fba':
        for rxn in mm.reactions:
            print(u'Flux\t{}\t{}'.format(rxn, result.get_value(_v(rxn))))
            if rxn in split_reversible_list:
                if '_reverse' in rxn:
                    rxn_f = rxn
                    rxn_f = rxn_f.replace('_reverse', '')
                    rxn_f += '_forward'
                    print(u'DGR\t{}\t{}'.format(
                        rxn, -1*result.get_value(_dgri(rxn_f))))
                else:
                    print(u'DGR\t{}\t{}'.format(
                        rxn, result.get_value(_dgri(rxn))))
            else:
                print(u'DGR\t{}\t{}'.format(rxn, result.get_value(_dgri(rxn))))
            print(u'Zi\t{}\t{}'.format(rxn, result.get_value(_zi(rxn))))
        for cp in cp_list:
            if cp not in excluded_compounds:
                print(u'CPD\t{}\t{}'.format(cp, result.get_value(_xij(cp))))
    elif simulation == 'l1min':
        _z = prob.namespace(mm.reactions, lower=0)
        z = _z.set(mm.reactions)
        v = _v.set(mm.reactions)
        prob.add_linear_constraints(z >= v, v >= -z)

        objective = _z.expr(
            (reaction_id, -1)
            for reaction_id in mm.reactions)
        prob.set_objective(objective)
        result = prob.solve_unchecked(lp.ObjectiveSense.Maximize)
        if not result.success:
            logger.error(u'Solution not optimal: {}'.format(result.status))
            quit()
        for rxn in mm.reactions:
            print(u'Flux\t{}\t{}'.format(rxn, result.get_value(_v(rxn))))
            if rxn in split_reversible_list:
                if '_reverse' in rxn:
                    rxn_f = rxn
                    rxn_f = rxn_f.replace('_reverse', '')
                    rxn_f += '_forward'
                    print(u'DGR\t{}\t{}'.format(
                        rxn, -1*result.get_value(_dgri(rxn_f))))
                else:
                    print(u'DGR\t{}\t{}'.format(
                        rxn, result.get_value(_dgri(rxn))))
            else:
                print(u'DGR\t{}\t{}'.format(
                    rxn, result.get_value(_dgri(rxn))))
            print(u'Zi\t{}\t{}'.format(rxn, result.get_value(_zi(rxn))))
        for cp in cp_list:
            if cp not in excluded_compounds:
                print(u'CPD\t{}\t{}'.format(cp, result.get_value(_xij(cp))))
    elif simulation == 'random':
        optimize = None
        for rxn_id in mm.reactions:
            if optimize is None:
                optimize = _v(rxn_id) * random.random()
            else:
                optimize += _v(rxn_id) * random.random()
        prob.add_linear_constraints(
            _v(objective) == result.get_value(_v(objective)))

        prob.set_objective(optimize)

        result = prob.solve_unchecked()
        for rxn in mm.reactions:
            print(u'Flux\t{}\t{}'.format(rxn, result.get_value(_v(rxn))))
            if rxn in split_reversible_list:
                if '_reverse' in rxn:
                    rxn_f = rxn
                    rxn_f = rxn_f.replace('_reverse', '')
                    rxn_f += '_forward'
                    print(u'DGR\t{}\t{}'.format(
                        rxn, -1*result.get_value(_dgri(rxn_f))))
                else:
                    print(u'DGR\t{}\t{}'.format(
                        rxn, result.get_value(_dgri(rxn))))
            else:
                print(u'DGR\t{}\t{}'.format(rxn, result.get_value(_dgri(rxn))))
            print(u'Zi\t{}\t{}'.format(rxn, result.get_value(_zi(rxn))))
        for cp in cp_list:
            if cp not in excluded_compounds:
                print(u'CPD\t{}\t{}'.format(cp, result.get_value(_xij(cp))))


def lump_parser(lump_file):
    """Parses a supplied file and returns
    dictionaries containing lump information.

    The supplied file should be in a tab separated table in the format of
    lumpID  lump_deltaG rxn1:1,rxn2:-1,rxn3:1  lumpRXN
    Returns dictionaries storing this information, linking reactions to the
    lumps.
    """
    rxn_to_lump_id = {}
    lump_to_rxn = {}
    lump_to_rxnids = {}
    lump_to_rxnids_dir = {}
    for row in csv.reader(lump_file, delimiter=str('\t')):
        lump_id, lump_rxn_list, lump_rxn = row
        rx_dir_list = []
        rx_list = []
        for i in lump_rxn_list.split(','):
            subrxn, dir = convert_to_unicode(i).split(':')
            rx_dir_list.append((subrxn, dir))
            rx_list.append(subrxn)
            rxn_to_lump_id[i] = lump_id
        lump_to_rxn[lump_id] = lump_rxn
        lump_to_rxnids[lump_id] = rx_list
        lump_to_rxnids_dir[lump_id] = rx_dir_list
    return lump_to_rxn, rxn_to_lump_id, lump_to_rxnids, lump_to_rxnids_dir


def parse_tparam_file(file):
    """Parse a transport parameter file.

    This file contains reaction IDs, net charge transported into
    the cell, and net protons transported into the cell.

    """
    t_param = {}
    if file is not None:
        for row in csv.reader(file, delimiter=str('\t')):
            rxn, c, h = row
            rxn = convert_to_unicode(rxn)
            t_param[rxn] = (Decimal(c), Decimal(h))
            t_param[u'{}_forward'.format(rxn)] = (Decimal(c), Decimal(h))
            t_param[u'{}_reverse'.format(rxn)] = (-Decimal(c), -Decimal(h))
    return t_param


def parse_dgf(mm, dgf_file):
    """A function that will parse a supplied deltaG of formation file.

    Compound IDs in this file do not need to contain the compartments.
    compound deltaGf values should be in Kcal/mol

    Args:
        mm: a metabolic model object
        dgf_file: a file that containing 2 columns of
            compound ids and deltaGf values
    """

    cpd_dgf_dict = {}
    for row in csv.reader(dgf_file, delimiter=str('\t')):
        for cpt in mm.compartments:
            try:
                dg = Decimal(row[1])
                derr = Decimal(row[2])
                cpd_dgf_dict[Compound(row[0], cpt)] = (dg, derr)
            except:
                logger.info(u'Compound {} has an assigned detltGf value '
                            'of {}. This is not an number and will be treated '
                            'as a missing '
                            'value.'.format(Compound(row[0], cpt), row[1]))
    return cpd_dgf_dict


def add_conc_constraints(xij, problem, cpd_conc_dict,
                         cp_list, water, hin, hout, hother):
    """Add concentration constraints to TMFA problem
    based on parsed concentration dictionary.

    """
    # Water and hydrogen need to be excluded from
    # these concentration constraints.
    excluded_cpd_list = []
    excluded_cpd_list.append(hin)
    excluded_cpd_list.append(hout)
    excluded_cpd_list.append(hother)
    for wat in water:
        excluded_cpd_list.append(wat)
    cpdid_xij_dict = {}
    for cp in cp_list:
        # define concentration variable for compound.
        cpdid_xij_dict[cp] = xij(cp)
        var = xij(cp)
        # Define default constraints for anything not set in the conc file
        if cp not in cpd_conc_dict.keys():
            if cp not in excluded_cpd_list:
                # Add concentration constraints as the
                # ln of the concentration (M).
                problem.add_linear_constraints(var >= math.log(0.00001))
                problem.add_linear_constraints(var <= math.log(0.02))
        elif cp in cpd_conc_dict.keys():
            if cp not in excluded_cpd_list:
                conc_limits = cpd_conc_dict[cp]
                if conc_limits[0] > conc_limits[1]:
                    logger.error(u'lower bound for {} concentration higher '
                                 'than upper bound'.format(cp))
                    quit()
                if Decimal(conc_limits[0]) == Decimal(conc_limits[1]):
                    problem.add_linear_constraints(
                        var == math.log(Decimal(conc_limits[0])))
                else:
                    problem.add_linear_constraints(
                        var >= math.log(Decimal(conc_limits[0])))
                    problem.add_linear_constraints(
                        var <= math.log(Decimal(conc_limits[1])))
    return problem, cpdid_xij_dict


def parse_dgr_file(dgr_file, mm):
    """Parses DeltaG of reaction file and returns a dictionary of the values.

    """
    def is_number(val):
        try:
            float(val)
            return True
        except ValueError:
            return False
    dgr_dict = {}
    for row in csv.reader(dgr_file, delimiter=str('\t')):
        rxn, dgr, err = row
        rxn = convert_to_unicode(rxn)
        if is_number(dgr):
            if is_number(err):
                err = Decimal(err)
            else:
                err = Decimal(2)
            if rxn in mm.reactions:
                dgr_dict[rxn] = (Decimal(dgr), err)
            elif '{}_forward'.format(rxn) in mm.reactions:
                dgr_dict[u'{}_forward'.format(rxn)] = (Decimal(dgr), err)
                dgr_dict[u'{}_reverse'.format(rxn)] = (-Decimal(dgr), err)
        else:
            logger.info(
                u'Reaction {} was provided with dgr value of {}'.format(
                    rxn, dgr))
    return dgr_dict


def calculate_dgr(mm, dgf_dict, excluded_reactions,
                  transport_parameters, ph_difference_rxn, scaled_compounds):
    """Calculates DeltaG values from DeltaG of formation values of compounds.

    """
    dgf_scaling = {}
    if scaled_compounds is not None:
        scaled_compounds.seek(0)
        for row in csv.reader(scaled_compounds, delimiter=str('\t')):
            dgf_scaling[row[0]] = Decimal(row[1])

    dgr_dict = {}
    for reaction in mm.reactions:
        if reaction not in excluded_reactions:
            dgr = 'NA'
            dgerr = 0
            rxn = mm.get_reaction(reaction)
            if any(dgf_dict.get(j[0]) is None for j in rxn.compounds):
                for j in rxn.compounds:
                    if j[0] not in dgf_dict.keys():
                        print(j[0])
                if reaction not in ph_difference_rxn:
                    logger.error(u'Reaction {} contains at least 1 compound '
                                 'with an unknown deltaGf '
                                 'value'.format(reaction))
                    quit()
            else:
                dgr = 0
                # Make a variable dgf_sum that represents
                # the sum of sij *  (stoichiometry *
                # deltaGf for reaction j.
                for cpd in rxn.compounds:
                    cpd_0 = convert_to_unicode(str(cpd[0]))
                    if cpd_0 in dgf_scaling.keys():
                        dgscale = dgf_scaling[cpd_0]
                    else:
                        dgscale = 1
                    (dg, dge) = dgf_dict[cpd[0].name]
                    dgs = Decimal(dg) * (Decimal(cpd[1])*dgscale)
                    dgr += dgs
                    dgerr += Decimal(cpd[1])*Decimal(dgscale) * dge
            dgr_dict[reaction] = (dgr, dgerr)
    return dgr_dict


def add_reaction_constraints(
        problem, _v, _zi, _dgri, _xij, mm, exclude_lumps, exclude_unknown,
        exclude_lumps_unknown, dgr_dict, lump_rxn_list, split_rxns,
        transport_parameters, testing_list, scaled_compounds, water, hin,
        hout, hother, ph, temp, err_est=False, hamilton=False):
    """Adds reaction constraints to a TMFA problem

    This function will add in gibbs free energy constraints to a TMFA flux
    problem. These constraints will be added based on provided Gibbs free
    energy values and temperature.

    Args:
        problem: :class:`psamm.lpsolver.lp.Problem`.
        mm: :class:`psamm.metabolicmodel.MetabolicModel`.
        _v: variable namespace for flux variables.
        _dgri: variable namespace for gibbs free energy variables
        _zi: variables namespace for indicator variables
        _xij: variable namespace for concentrations
        exclude_unknown: List of reactions excluded from thermo constraints.
        exclude_lumps_unknown: List of excluded reactions and lumped reactions.
        dgr_dict: dictionary where keys are reaction ids and values
            are deltag values.
        lump_rxn_list: List of lump reaction IDs.
        split_rxns: List of tuples of (reaction_forward, reaction_reverse)
        transport_parameters: dictionary of reaction IDs to proton
            transport and charge transport values.
        testing_list: List of reactions to add deltaG constraints for.
        water: list of water compound IDs.
        hin: ID of proton compound from inside cell compartment
        hout: ID of proton compound from outside cell compartment
        hother: List of other proton IDs
        ph: list of two tuples containing pH bounds. [(in_lower, in_upper),
            (out_lower, out_upper)]
        temp: Temperature in Celsius
        err_est: True or False for using error estimates for deltaG values.
        hamilton: True or False for using Hamilton TMFA method.

    """
    dgf_scaling = {}
    if scaled_compounds is not None:
        scaled_compounds.seek(0)
        for row in csv.reader(scaled_compounds, delimiter=str('\t')):
            dgf_scaling[row[0]] = Decimal(row[1])

    # "idg" is the ideal gas constant in units of kJ/mol
    idg = Decimal(8.3144621 / 1000)  # kJ/mol
    # R = Decimal(1.987 / 1000) kcal/mol

    # "tkelvin" is the temperature on the Kelvin scale
    tkelvin = Decimal(temp) + Decimal(273.15)

    k = 500
    epsilon = 0.000001
    ph_in = ph[0]
    ph_out = ph[0]
    h_p = _xij(str(hout))

    problem.add_linear_constraints(h_p <= ph_out[1])
    problem.add_linear_constraints(h_p >= ph_out[0])

    h_c = _xij(str(hin))

    problem.add_linear_constraints(h_c >= ph_in[0])
    problem.add_linear_constraints(h_c <= ph_in[1])
    delta_ph = (h_c - h_p)

    # "fc" is the Faraday constant in units of kcal/(mV*mol)
    fc = Decimal(0.02306)

    excluded_cpd_list = []
    excluded_cpd_list.append(hin)
    excluded_cpd_list.append(hout)
    if hother is not None:
        excluded_cpd_list.append(hother)
    for wat in water:
        excluded_cpd_list.append(wat)
    logger.info(u'Excluded compounds: {}'.format(','.join(excluded_cpd_list)))
    logger.info(u'Temperature: {}'.format(tkelvin))
    logger.info(u'using h in {}'.format(hin))
    logger.info(u'using h out {}'.format(hout))
    logger.info(u'using water {}'.format(water))
    logger.info(u'using ph range of {} to {} for internal compartment'.format(
        ph_in[0], ph_in[1]))
    logger.info(u'using ph range of {} to {} for external compartment'.format(
        ph_out[0], ph_out[1]))
    # for (f, r) in split_rxns:
    #     split_list.append(f)
    #     split_list.append(r)
    #     if f not in exclude_unknown:
    #         dgrif = _dgri(f)
    #         dgrir = _dgri(r)
    #         problem.add_linear_constraints(dgrif == dgrir * -1)

    new_excluded_reactions = []
    for reaction in mm.reactions:
        if reaction not in exclude_unknown:
            rxn = mm.get_reaction(reaction)
            rhs_check = 0
            lhs_check = 0
            for (cpd, stoich) in rxn.compounds:
                if stoich < 0:
                    if str(cpd) not in excluded_cpd_list:
                        lhs_check += 1
                if stoich > 0:
                    if str(cpd) not in excluded_cpd_list:
                        rhs_check += 1
            if rhs_check == 0 or lhs_check == 0:
                new_excluded_reactions.append(reaction)

            zi = _zi(reaction)
            split_rxns_l = []
            for (f, r) in split_rxns:
                split_rxns_l.append(f)
                split_rxns_l.append(r)
            if reaction in split_rxns_l:
                if '_reverse' in reaction:
                    f_rxn = reaction.replace('_reverse', '')
                    f_rxn = f_rxn+'_forward'
                    dgri = -1*_dgri(reaction)
                else:
                    dgri = _dgri(reaction)
            else:
                dgri = _dgri(reaction)
            vi = _v(reaction)
            vmax = mm.limits[reaction].upper
            if reaction in testing_list:
                if reaction in transport_parameters.keys():
                    (c, h) = transport_parameters[reaction]
                    ddph = Decimal(-2.3)*Decimal(h)*idg*tkelvin*delta_ph
                    dpsi = Decimal(33.33) * delta_ph - Decimal(143.33)
                    ddpsi = dpsi * Decimal(c) * Decimal(fc)
                    dgr_trans = ddph + ddpsi
                else:
                    dgr_trans = 0
                (dgr0, err) = dgr_dict[reaction]
                ssxi = 0

                if err_est:
                    problem.define(u'dgr_err_{}'.format(reaction),
                                   types=lp.VariableType.Continuous,
                                   lower=-1000, upper=1000)
                    dgr_err = problem.var(u'dgr_err_{}'.format(reaction))
                    problem.add_linear_constraints(dgr_err <= 2*err)
                    problem.add_linear_constraints(dgr_err >= -2*err)
                else:
                    dgr_err = 0

                for (cpd, stoich) in rxn.compounds:
                    cpd = convert_to_unicode(str(cpd))
                    if cpd not in excluded_cpd_list:
                        scale = dgf_scaling.get(cpd, 1)
                        ssxi += _xij(cpd) * Decimal(stoich) * scale

                problem.add_linear_constraints(
                    dgri == dgr0 + (idg * tkelvin * ssxi
                                    ) + dgr_err + dgr_trans)
                if hamilton:
                    problem.add_linear_constraints(dgri <= 300-epsilon)
                    problem.add_linear_constraints(dgri >= -300+epsilon)

                if reaction not in exclude_lumps_unknown:
                    if rhs_check != 0 and lhs_check != 0:
                        problem.add_linear_constraints(
                            dgri - k + (k * zi) <= -epsilon)
                        problem.add_linear_constraints(vi <= zi * vmax)

        if reaction in lump_rxn_list.keys():
            problem.define(u'yi_{}'.format(reaction), lower=int(0),
                           upper=int(1), types=lp.VariableType.Binary)
            if reaction not in new_excluded_reactions:
                vi = _v(reaction)
                yi = problem.var(u'yi_{}'.format(reaction))
                dgri = _dgri(reaction)
                problem.add_linear_constraints(vi == 0)
                problem.add_linear_constraints(dgri - (k * yi) <= - epsilon)
                sub_rxn_list = lump_rxn_list[reaction]
                sszi = 0
                for sub_rxn in sub_rxn_list:
                    sszi += _zi(sub_rxn)
                problem.add_linear_constraints(
                    yi + sszi <= len(sub_rxn_list))

    for (forward, reverse) in split_rxns:
        problem.add_linear_constraints(
            _zi(forward) + _zi(reverse) <= int(1))
    return problem, excluded_cpd_list



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals, division

import time
import logging

from ..command import (Command, MetabolicMixin, ObjectiveMixin,
                       SolverCommandMixin, FilePrefixAppendAction)
from .. import fluxanalysis
from .. import moma
from ..expression import boolean

from six import string_types, text_type

logger = logging.getLogger(__name__)


class GeneDeletionCommand(MetabolicMixin, ObjectiveMixin, SolverCommandMixin,
                          Command):
    """Find reactions requiring a specified gene and delete them from model.

    Reports the new objective flux after the gene was deleted.
    """
    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--gene', metavar='genes', action=FilePrefixAppendAction,
            type=str, default=[], help='Delete multiple genes from model')
        parser.add_argument(
            '--method', metavar='method',
            choices=['fba', 'lin_moma', 'lin_moma2', 'moma', 'moma2'],
            type=text_type, default='fba',
            help='Select which method to use. (fba, lin_moma, lin_moma2, '
                 'moma, moma2)')
        super(GeneDeletionCommand, cls).init_parser(parser)

    def run(self):
        """Delete the specified gene and solve using the desired method."""
        obj_reaction = self._get_objective()

        genes = set()
        gene_assoc = {}
        for reaction in self._model.reactions:
            assoc = None
            if reaction.genes is None:
                continue
            elif isinstance(reaction.genes, string_types):
                assoc = boolean.Expression(reaction.genes)
            else:
                variables = [boolean.Variable(g) for g in reaction.genes]
                assoc = boolean.Expression(boolean.And(*variables))
            genes.update(v.symbol for v in assoc.variables)
            gene_assoc[reaction.id] = assoc

        reactions = set(self._mm.reactions)
        start_time = time.time()
        testing_genes = set(self._args.gene)
        deleted_reactions = set()

        logger.info('Trying model without genes: {}...'.format(
                    ', '.join(sorted(testing_genes))))

        for reaction in reactions:
            if reaction not in gene_assoc:
                continue
            assoc = gene_assoc[reaction]
            if any(boolean.Variable(gene) in assoc.variables
                    for gene in testing_genes):
                new_assoc = assoc.substitute(
                    lambda v: v if v.symbol not in testing_genes else False)
                if new_assoc.has_value() and not new_assoc.value:
                    logger.info('Deleting reaction {}...'.format(reaction))
                    deleted_reactions.add(reaction)

        if self._args.method in ['moma', 'moma2']:
            solver = self._get_solver(quadratic=True)
        else:
            solver = self._get_solver()

        if self._args.method == 'fba':
            logger.info('Solving using FBA...')
            prob = fluxanalysis.FluxBalanceProblem(self._mm, solver)

            try:
                prob.maximize(obj_reaction)
            except fluxanalysis.FluxBalanceError as e:
                self.report_flux_balance_error(e)

            wild = prob.get_flux(obj_reaction)

            for reaction in deleted_reactions:
                flux_var = prob.get_flux_var(reaction)
                prob.prob.add_linear_constraints(flux_var == 0)

            prob.maximize(obj_reaction)
            deleteflux = prob.get_flux(obj_reaction)
        elif self._args.method in ['lin_moma', 'lin_moma2', 'moma', 'moma2']:
            prob = moma.MOMAProblem(self._mm, solver)
            wt_fluxes = prob.get_minimal_fba_flux(obj_reaction)
            wild = wt_fluxes[obj_reaction]

            for reaction in deleted_reactions:
                flux_var = prob.get_flux_var(reaction)
                prob.prob.add_linear_constraints(flux_var == 0)

            try:
                if self._args.method == 'moma':
                    logger.info('Solving using MOMA...')
                    prob.moma(wt_fluxes)
                elif self._args.method == 'lin_moma':
                    logger.info('Solving using linear MOMA...')
                    prob.lin_moma(wt_fluxes)
                elif self._args.method == 'moma2':
                    logger.info('Solving using combined-model MOMA...')
                    prob.moma2(obj_reaction, wild)
                elif self._args.method == 'lin_moma2':
                    logger.info('Solving using combined-model linear MOMA...')
                    prob.lin_moma2(obj_reaction, wild)
            except moma.MOMAError:
                self.fail('Error computing the MOMA result.')

            deleteflux = prob.get_flux(obj_reaction)

        logger.info(
            'Solving took {:.2f} seconds'.format(time.time() - start_time))
        logger.info(
            'Objective reaction after gene deletion has flux {}'.format(
                deleteflux + 0))
        if wild != 0:
            logger.info(
                'Objective reaction has {:.2%} flux of wild type flux'.format(
                    abs(deleteflux / wild)))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2018-2020  Jing Wang <wjingsjtu@gmail.com>
# Copyright 2020-2021  Elysha Sameth <esameth1@uri.edu>

# -*- coding: utf-8 -*-

"""Generate a common model for two distinct metabolic models."""

from __future__ import print_function, division

from builtins import object, input
import operator
from itertools import product
import time
import sys

import psamm.bayesian as bayesian
import psamm.translate_id as tr_id
import psamm.manual_curation as curation
from psamm.util import mkdir_p
from psamm.datasource.native import ModelReader
from psamm.command import Command
from psamm.datasource.reaction import Compound


class ModelMappingCommand(Command):
    """Generate a common model for two distinct metabolic models."""

    @classmethod
    def init_parser(cls, parser):
        """Initialize argument parser"""
        subparsers = parser.add_subparsers(title="Tools")

        # model mapping subcommand
        parser_mm = subparsers.add_parser(
            'map', help='Bayesian model mapping')
        parser_mm.set_defaults(which='mm')
        parser_mm.add_argument(
            '--dest-model', type=str, required=True, help='Model to map to')
        parser_mm.add_argument(
            '--compound-map', type=str, nargs='?',
            help='Actual compound mapping, evaluation only')
        parser_mm.add_argument(
            '--reaction-map', type=str, nargs='?',
            help='Actual reaction mapping, evaluation only')
        parser_mm.add_argument(
            '-c', '--consistency-check', action='store_true',
            help=(
                'Check the consistency of compound id '
                'between compounds.yaml and reactions.yaml'))
        parser_mm.add_argument(
            '-n', '--nproc', type=int, action='store', default=1,
            help='Number of processes to use (default = 1)')
        parser_mm.add_argument(
            '-o', '--outpath', type=str,
            action='store', default='.',
            help=(
                'Path to store output files '
                '(default = current position)'))
        parser_mm.add_argument(
            '--threshold-compound', type=float, default=0,
            help=(
                'likelihood threshold for compound mapping '
                '(default = 0)'))
        parser_mm.add_argument(
            '--threshold-reaction', type=float, default=0,
            help=(
                'likelihood threshold for reaction mapping '
                '(default = 0)'))
        parser_mm.add_argument(
            '--map-compound-kegg', action='store_true',
            help='Compare KEGG id during compound mapping')
        parser_mm.add_argument(
            '--map-reaction-gene', action='store_true',
            help='Compare gene list during reaction mapping')
        parser_mm.add_argument(
            '--raw', action='store_true',
            help=(
                'Output raw pairwise mapping results as well, '
                'may be very big with large models'))
        parser_mm.add_argument(
            '--log', action='store_true',
            help=(
                'Output log file of the p_match and p_no_match '
                'for each feature, '
                'may be very big with large models'))
        parser_mm.add_argument(
            '--compartment-map-file', type=str,
            help=(
                'A tab delimited file (.tsv) listing the compartment ids of '
                'query model in the first column, and the corresponding '
                'compartment ids of dest model in the second column. Required '
                'if the compartment ids in the two models are not consistent.'
            )
        )
        parser_mm.add_argument(
            '--gene-map-file', type=str,
            help=(
                'A tab delimited file (.tsv) listing the gene ids of '
                'query model in the first column, and the corresponding '
                'gene ids of dest model in the second column. Required '
                'if the gene ids in the two models are not consistent.'
            )
        )

        # manual curation subcommand
        parser_c = subparsers.add_parser(
            'manual_curation',
            help='Interactive mode to check mapping result')
        parser_c.set_defaults(which='curation')
        parser_c.add_argument(
            '--dest-model', type=str, required=True, help='Model to map to')
        parser_c.add_argument(
            '--compound-map', type=str, required=True,
            help='Compound model mapping result')
        parser_c.add_argument(
            '--reaction-map', type=str, required=True,
            help='Reaction model mapping result')
        parser_c.add_argument(
            '--curated-compound-map', type=str, required=True,
            help=(
                'File to store curated compound mapping, if file exists, '
                'resume previous curation'))
        parser_c.add_argument(
            '--curated-reaction-map', type=str, required=True,
            help=(
                'File to store curated reaction mapping, if file exists, '
                'resume previous curation'))
        parser_c.add_argument(
            '--compartment-map', type=str,
            help=(
                'A tab delimited file (.tsv) listing the compartment ids of '
                'query model in the first column, and the corresponding '
                'compartment ids of dest model in the second column. Required '
                'if the compartment ids in the two models are not consistent.'
            )
        )
        parser_c.add_argument(
            '--compound-only', action='store_true',
            help='Set to only curate compound maps.')

        # translate id subcommand
        parser_t = subparsers.add_parser(
            'translate_id', help='Translate ids based on mapping result')
        parser_t.set_defaults(which='translateid')
        parser_t.add_argument(
            '--compound-map', type=str, required=True,
            help=(
                'Tab-delimited table, the first two columns store the '
                'original and target ids'))
        parser_t.add_argument(
            '--reaction-map', type=str, required=True,
            help=(
                'Tab-delimited table, the first two columns store the '
                'original and target ids'))
        parser_t.add_argument(
            '--compartment-map', type=str, required=False,
            help=(
                'Tab-delimited table, the first two columns store the '
                'original and target ids'))
        parser_t.add_argument(
            '-o', '--outpath', type=str,
            action='store', default='.',
            help=(
                'Path to store output files '
                '(default = current position)'))

    def run(self):
        """Run model mapping command."""

        which_command = self._args.which
        if which_command == 'mm':
            self._model_mapping()
        if which_command == 'translateid':
            self._translate_id()
        if which_command == 'curation':
            self._curation()

    def _model_mapping(self):
        """Run model mapping"""
        # Parse models
        model1 = ModelReader.reader_from_path(
            self._args.model).create_model()
        model2 = ModelReader.reader_from_path(
            self._args.dest_model).create_model()

        # Check if any compound charge is not integer
        invalid_cpd = []
        for compound in model1.compounds:
            result = bayesian.check_cpd_charge(compound, model1.name)
            if not result:
                invalid_cpd.append(compound.id)
        for compound in model2.compounds:
            result = bayesian.check_cpd_charge(compound, model2.name)
            if not result:
                invalid_cpd.append(compound.id)
        if len(invalid_cpd) != 0:
            quit()

        # Read model into internal format
        model1 = bayesian.MappingModel(model1)
        model2 = bayesian.MappingModel(model2)

        # Model summaries
        model1.print_summary()
        print('\n')
        model2.print_summary()
        print('\n')

        # Load actual model mappings
        actual_compound_mapping = None
        if self._args.compound_map is not None:
            with open(self._args.compound_map, 'r') as f:
                actual_compound_mapping = dict(read_mapping_file(f))

        actual_reaction_mapping = None
        if self._args.reaction_map is not None:
            with open(self._args.reaction_map, 'r') as f:
                actual_reaction_mapping = dict(read_mapping_file(f))

        # Load compartment mapping
        compartment_map = {}
        if self._args.compartment_map_file is not None:
            with open(self._args.compartment_map_file, 'r') as f:
                for row in f:
                    source, target = row.strip().split()
                    compartment_map[source] = target

        # Load gene mapping
        gene_map = {}
        if self._args.gene_map_file is not None:
            with open(self._args.gene_map_file, 'r') as f:
                for row in f:
                    source, target = row.strip().split()
                    gene_map[source] = target

        # Check models
        if (not (model1.check_reaction_compounds() and
                 model2.check_reaction_compounds())):
            quit((
                '\nError: '
                'equations have something not listed in compounds.yaml, '
                'please check it!'))

        mkdir_p(self._args.outpath)

        if actual_compound_mapping is not None:
            mkdir_p(self._args.outpath + '/roc')

        # Bayesian classifier
        print('Using %i processes...' % (self._args.nproc))
        t = time.time()
        cpd_bayes_pred = bayesian.BayesianCompoundPredictor(
            model1, model2, self._args.nproc, self._args.outpath,
            log=self._args.log, kegg=self._args.map_compound_kegg)
        print(
            'It took %.2f seconds to calculate compound mapping...'
            % (time.time() - t))

        print('Writing output...')
        sys.stdout.flush()
        t = time.time()
        # Write out ROC curve results
        if actual_compound_mapping is not None:
            with open(self._args.outpath +
                      '/roc/compound_bayes.tsv', 'w') as f:
                write_roc_curve(f, cpd_bayes_pred.model1.compounds,
                                cpd_bayes_pred.model2.compounds,
                                cpd_bayes_pred,
                                actual_compound_mapping)

        # Parse and output raw mapping
        if self._args.raw:
            cpd_bayes_pred.get_raw_map().to_csv(
                self._args.outpath + '/bayes_compounds.tsv', sep='\t',
                encoding='utf-8')

        # Output best mapping
        compound_best = cpd_bayes_pred.get_best_map(
            self._args.threshold_compound)
        compound_best.to_csv(
            self._args.outpath + '/bayes_compounds_best.tsv', sep='\t',
            encoding='utf-8')
        print('It took %.2f seconds to write output...' % (time.time() - t))
        sys.stdout.flush()

        # Bayesian classifier
        t = time.time()
        rxn_bayes_pred = bayesian.BayesianReactionPredictor(
            model1, model2, compound_best.loc[:, 'p'],
            self._args.nproc, self._args.outpath,
            log=self._args.log, gene=self._args.map_reaction_gene,
            compartment_map=compartment_map, gene_map=gene_map)
        print(
            'It took %.2f seconds to calculate reaction mapping...'
            % (time.time() - t))

        print('Writing output...')
        sys.stdout.flush()
        t = time.time()
        # Write out ROC curve results
        if actual_reaction_mapping is not None:
            with open(self._args.outpath +
                      '/roc/reaction_bayes.tsv', 'w') as f:
                write_roc_curve(f, rxn_bayes_pred.model1.reactions,
                                rxn_bayes_pred.model2.reactions,
                                rxn_bayes_pred,
                                actual_reaction_mapping)

        # Parse and output raw mapping
        if self._args.raw:
            rxn_bayes_pred.get_raw_map().to_csv(
                self._args.outpath + '/bayes_reactions.tsv', sep='\t',
                encoding='utf-8')

        # Output best mapping
        reaction_best = rxn_bayes_pred.get_best_map(
            self._args.threshold_reaction)
        reaction_best.to_csv(
            self._args.outpath + '/bayes_reactions_best.tsv', sep='\t',
            encoding='utf-8')
        print('It took %.2f seconds to write output...' % (time.time() - t))
        sys.stdout.flush()

    def _curation(self):
        # read models
        model1 = ModelReader.reader_from_path(self._args.model)
        model1 = model1.create_model()
        model2 = ModelReader.reader_from_path(self._args.dest_model)
        model2 = model2.create_model()

        # initiate curator
        curator = curation.Curator(
            self._args.compound_map, self._args.reaction_map,
            self._args.curated_compound_map, self._args.curated_reaction_map)

        # read compartment map
        compartment_map = {}
        if self._args.compartment_map is not None:
            with open(self._args.compartment_map) as f:
                for row in f:
                    old, new = row.strip().split()
                    compartment_map[old] = new

        if self._args.compound_only:
            self._curate_compound(curator, model1, model2)
        else:
            self._curate_reaction(curator, model1, model2, compartment_map)

    def _curate_reaction(self, curator, model1, model2, compartment_map):
        print('Starting to do manual curation...\n')
        print('Type "stop" to stop\n')
        # iterate through reaction mapping
        for rmap in curator.reaction_map.iterrows():
            # skip already curated reactions
            if (curator.reaction_checked(rmap[0]) or
                    curator.reaction_ignored(rmap[0][0])):
                continue
            print('You have mapped %i reactions. '
                  'There are %i reactions unmapped.'
                  % (curator.num_curated_reactions,
                     curator.num_curated_reactions_left))
            print('\n')
            # check the compound mapping in current reaction
            compounds = curation.search_reaction(model1, [rmap[0][0]])
            compounds = [c.name for c in next(compounds)]
            dest_compounds = curation.search_reaction(model2, [rmap[0][1]])
            dest_compounds = [c.name for c in next(dest_compounds)]
            print('\n', '*' * 60, '\n')
            print('Below are the compound mapping involved:\n')
            for compound in compounds:
                for cmap in curator.compound_map.loc[compound].iterrows():
                    if (cmap is None or
                            cmap[0] not in dest_compounds or
                            curator.compound_checked((compound, cmap[0])) or
                            curator.compound_ignored(compound)):
                        continue
                    print('You have mapped %i compounds. '
                          'There are %i compounds unmapped.'
                          % (curator.num_curated_compounds,
                             curator.num_curated_compounds_left))
                    print('\n')
                    print('-' * 30)
                    print(cmap[1])
                    print('-' * 30)
                    print('\n')
                    curation.search_compound(model1, [compound])
                    curation.search_compound(model2, [cmap[0]])
                    # waiting for legal curation input
                    while True:
                        ask = input(
                            ('True compound match? (y/n/ignore/save/stop, '
                             'type ignore to ignore this compound in future, '
                             'type save to save current progress, '
                             'type stop to save and exit): '))
                        if ask.lower() == 'n':
                            curator.add_mapping(
                                (compound, cmap[0]),
                                'c',
                                False
                            )
                            break
                        if ask.lower() == 'y':
                            curator.add_mapping(
                                (compound, cmap[0]),
                                'c',
                                True
                            )
                            curator.num_curated_compounds += 1
                            curator.num_curated_compounds_left -= 1
                            break
                        if ask.lower() == 'ignore':
                            curator.add_ignore(compound, 'c')
                            break
                        if ask.lower() == 'stop':
                            curator.save()
                            exit(0)
                        if ask.lower() == 'save':
                            curator.save()

                    print('\n')
            print('You have mapped %i reactions. '
                  'There are %i reactions unmapped.'
                  % (curator.num_curated_reactions,
                     curator.num_curated_reactions_left))
            print('\n')
            print('Here is the reaction mapping:')
            print('-' * 30)
            print(rmap[1])
            print('-' * 30)
            print('\n')
            compounds = curation.search_reaction(model1, [rmap[0][0]])
            compounds = [c for c in next(compounds)]
            dest_compounds = curation.search_reaction(model2, [rmap[0][1]])
            dest_compounds = [c for c in next(dest_compounds)]
            print(('These two reactions have the following curated compound '
                   'pairs:\n'))
            count = 0
            compounds_curated = set()
            dest_compounds_curated = set()
            for compound in compounds:
                if compound.name in curator.curated_compound_map.index:
                    for cmap in curator.curated_compound_map.loc[
                            compound.name].iterrows():
                        compartment = compartment_map.get(
                            compound.compartment, compound.compartment
                        )
                        dest_compound = Compound(cmap[0], compartment)
                        if dest_compound in dest_compounds:
                            print(compound, dest_compound, cmap[1]['p'])
                            compounds_curated.add(compound)
                            dest_compounds_curated.add(dest_compound)
                            count += 1
            print('\n')
            print('%i compounds in %s' % (len(compounds), rmap[0][0]))
            unpaired = set(compounds) - compounds_curated
            if len(unpaired) > 0:
                print('Unpaired compounds: %s\n' % ', '.join(
                    [str(c) for c in unpaired]))
            print('%i compounds in %s' % (len(dest_compounds), rmap[0][1]))
            unpaired = set(dest_compounds) - dest_compounds_curated
            if len(unpaired) > 0:
                print('Unpaired compounds: %s\n' % ', '.join(
                    [str(c) for c in unpaired]))
            print('%i curated compound pairs\n' % count)
            ask = ''
            # waiting for legal curation input
            while ask not in ['y', 'n', 'stop']:
                ask = input(
                    ('True reaction match? (y/n/ignore/save/stop, '
                     'type ignore to ignore this reaction in future, '
                     'type save to save current progress, '
                     'type stop to save and exit): ')
                )
                if ask.lower() == 'n':
                    curator.add_mapping(rmap[0], 'r', False)
                    break
                if ask.lower() == 'y':
                    curator.add_mapping(rmap[0], 'r', True)
                    curator.num_curated_reactions += 1
                    curator.num_curated_reactions_left -= 1
                    break
                if ask.lower() == 'ignore':
                    curator.add_ignore(rmap[0][0], 'r')
                    break
                if ask.lower() == 'stop':
                    curator.save()
                    exit(0)
                if ask.lower() == 'save':
                    curator.save()
            print('\n', '=' * 60, '\n')
        curator.save()

    def _curate_compound(self, curator, model1, model2):
        for cmap in curator.compound_map.iterrows():
            if (curator.compound_checked(cmap[0]) or
                    curator.compound_ignored(cmap[0][0])):
                continue
            print('You have mapped %i compounds. '
                  'There are %i compounds unmapped.'
                  % (curator.num_curated_compounds,
                     curator.num_curated_compounds_left))
            print('-' * 30)
            print(cmap[1])
            print('-' * 30)
            print('\n')
            curation.search_compound(model1, [cmap[0][0]])
            curation.search_compound(model2, [cmap[0][1]])
            # waiting for legal curation input
            while True:
                ask = input(
                    ('True compound match? (y/n/ignore/save/stop, '
                     'type ignore to ignore this compound in future, '
                     'type save to save current progress, '
                     'type stop to save and exit): '))
                if ask.lower() == 'n':
                    curator.add_mapping(
                        cmap[0],
                        'c',
                        False
                    )
                    break
                if ask.lower() == 'y':
                    curator.add_mapping(
                        cmap[0],
                        'c',
                        True
                    )
                    curator.num_curated_compounds += 1
                    curator.num_curated_compounds_left -= 1
                    break
                if ask.lower() == 'ignore':
                    curator.add_ignore(cmap[0][0], 'c')
                    break
                if ask.lower() == 'stop':
                    curator.save()
                    exit(0)
                if ask.lower() == 'save':
                    curator.save()
            print('\n', '=' * 60, '\n')

    def _translate_id(self):
        """Translate ids"""
        # make output dir
        mkdir_p(self._args.outpath)

        # read mapping files
        cpd_mapping_id = tr_id.read_mapping(self._args.compound_map)
        rxn_mapping_id = tr_id.read_mapping(self._args.reaction_map)
        if self._args.compartment_map is not None:
            compartment_mapping_id = tr_id.read_mapping(
                self._args.compartment_map)
        else:
            compartment_mapping_id = None

        # make output model
        out_nm = tr_id.TranslatedModel(
            self._model,
            cpd_mapping_id,
            rxn_mapping_id,
            compartment_mapping_id
        )

        out_nm.write_model(dest=self._args.outpath, split_subsystem=False)


def read_mapping_file(f):
    """Read a mapping file and yield source, target-set tuples"""
    for line in f:
        source, target = line.strip().split(None, 1)
        if target == '-':
            target_set = None
        else:
            target_set = set(s for s in target.split(','))
        yield source, target_set


class Confusion(object):
    """Confusion matrix."""

    def __init__(self, tp=0, tn=0, fp=0, fn=0):
        self.tp, self.tn, self.fp, self.fn = tp, tn, fp, fn

    @property
    def positive(self):
        return self.tp + self.fn

    @property
    def negative(self):
        return self.fp + self.tn

    @property
    def total(self):
        return self.tp + self.tn + self.fp + self.fn

    @property
    def tp_rate(self):
        return self.tp if self.tp == 0 else self.tp / self.positive

    @property
    def fp_rate(self):
        return self.fp if self.fp == 0 else self.fp / self.negative

    @property
    def specificity(self):
        return self.tn if self.tn == 0 else self.tn / self.negative

    @property
    def precision(self):
        return self.tp if self.tp == 0 else self.tp / (self.tp + self.fp)

    @property
    def accuracy(self):
        return (self.tp + self.tn) / self.total

    @property
    def balanced_accuracy(self):
        return (self.tp_rate + self.specificity) / 2.0

    @property
    def informedness(self):
        return self.tp_rate + self.specificity - 1.0

    def __repr__(self):
        return 'Confusion(tp={}, tn={}, fp={}, fn={})'.format(
            self.tp, self.tn, self.fp, self.fn)


def evaluate_roc_points(model1, model2, predictor, actual):
    """Return confusion matrix for each point on the ROC curve"""

    # Count total positive and negative pairs
    positive, negative = 0, 0
    for e1, e2 in product(model1, model2):
        correct = e1 in actual and actual[e1] is not None and e2 in actual[e1]
        positive += correct
        negative += not correct

    tp, fp = 0, 0
    p_prev = None
    for pair, p in sorted((((e1, e2), predictor.map(e1, e2))
                           for e1, e2 in product(model1, model2)),
                          key=operator.itemgetter(1), reverse=True):
        e1, e2 = pair
        if p != p_prev:
            yield p, Confusion(
                tp=tp, fp=fp, fn=positive - tp, tn=negative - fp)
            p_prev = p
        correct = e1 in actual and actual[e1] is not None and e2 in actual[e1]
        tp += correct
        fp += not correct

    cf = Confusion(tp=tp, fp=fp, fn=positive - tp, tn=negative - fp)
    yield -float('inf'), cf


def write_roc_curve(out_file, model1, model2, predictor, actual):
    """Calculate ROC points and write them to file"""
    for p, conf in evaluate_roc_points(model1, model2,
                                       predictor, actual):
        out_file.write('{}\t{}\t{}\t{}\t{}\n'.format(
            p, conf.tp, conf.tn, conf.fp, conf.fn))



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2016-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2016-2017  Brian Bishop <brian_bishop@my.uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Implementation of Minimization of Metabolic Adjustments (MOMA)."""

import logging
from itertools import product

from six import iteritems, text_type, raise_from

from psamm.lpsolver import lp

logger = logging.getLogger(__name__)


class MOMAError(Exception):
    """Error indicating an error solving MOMA."""


class ConstraintGroup(object):
    """Constraints that will be imposed on the model when solving.

    Args:
        moma: MOMAProblem object for the proposed constraints.
        *args: The constraints that are imposed on the model.
    """
    def __init__(self, moma, *args):
        self._moma = moma
        self._constrs = []
        if len(args) > 0:
            self.add(*args)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.delete()

    def add(self, *args):
        """Add constraints to the model."""
        self._constrs.extend(self._moma._prob.add_linear_constraints(*args))

    def delete(self):
        """Set up the constraints to get deleted on the next solve."""
        self._moma._remove_constr.extend(self._constrs)


class MOMAProblem(object):
    """Model as a flux optimization problem with minimal flux redistribution.

    Create a representation of the model as an LP optimization problem with
    steady state assumption and a minimal redistribution of metabolic fluxes
    with respect to the wild type configuration.

    The problem can be solved using any of the four MOMA variants described in
    [Segre02]_ and [Mo09]_. MOMA is formulated to avoid the FBA assumption that
    that growth efficiency has evolved to an optimal point directly following
    model perturbation. MOMA finds the optimal solution for a model with
    minimal flux redistribution with respect to the wild type flux
    configuration.

    MOMA is implemented with two variations of a linear optimization problem
    (:meth:`.lin_moma` and :meth:`.lin_moma2`) and two variations of a
    quadratic optimization problem (:meth:`.moma` and :meth:`.moma2`).
    Further information on these methods can be found within their respective
    documentation.

    The problem can be modified and solved as many times as needed. The flux
    of a reaction can be obtained after solving using :meth:`.get_flux`.

    Args:
        model: :class:`MetabolicModel` to solve.
        solver: LP solver instance to use.
    """
    def __init__(self, model, solver):
        self._prob = solver.create_problem()
        self._model = model

        self._remove_constr = []

        self._v_wt = v_wt = self._prob.namespace(name='v_wt')
        self._v = v = self._prob.namespace(name='v')
        self._z = z = self._prob.namespace(name='z')
        self._z_diff = z_diff = self._prob.namespace(name='z_diff')

        # Define flux variables
        for reaction_id in self._model.reactions:
            lower, upper = self._model.limits[reaction_id]
            v_wt.define([reaction_id], lower=lower, upper=upper)
            v.define([reaction_id], lower=lower, upper=upper)
            z.define([reaction_id], lower=0, upper=max(abs(lower), abs(upper)))

            flux_range = float(upper) - float(lower)
            z_diff.define([reaction_id], lower=0, upper=flux_range)

        # Define constraints
        mass_balance = self.constraints()
        massbalance_lhs = {
            spec: 0 for spec in product(model.compounds, ('wt', 'mod'))}
        for (compound, reaction_id), value in iteritems(self._model.matrix):
            massbalance_lhs[compound, 'wt'] += v_wt[reaction_id] * value
            massbalance_lhs[compound, 'mod'] += v[reaction_id] * value
        for compound, lhs in iteritems(massbalance_lhs):
            mass_balance.add(lhs == 0)

    @property
    def prob(self):
        """Return the underlying LP problem."""
        return self._prob

    def constraints(self, *args):
        """Return a constraint object."""
        return ConstraintGroup(self, *args)

    def _adjustment_reactions(self):
        """Yield all the non exchange reactions in the model."""
        for reaction_id in self._model.reactions:
            if not self._model.is_exchange(reaction_id):
                yield reaction_id

    def _solve(self, sense=None):
        """Remove old constraints and then solve the current problem.

        Args:
            sense: Minimize or maximize the objective.
                (:class:`.lp.ObjectiveSense)

        Returns:
            The Result object for the solved LP problem
        """
        # Remove the constraints from the last run
        while len(self._remove_constr) > 0:
            self._remove_constr.pop().delete()

        try:
            return self._prob.solve(sense=sense)
        except lp.SolverError as e:
            raise_from(MOMAError(text_type(e)), e)
        finally:
            self._remove_constr = []

    def solve_fba(self, objective):
        """Solve the wild type problem using FBA.

        Args:
            objective: The objective reaction to be maximized.

        Returns:
            The LP Result object for the solved FBA problem.
        """
        self._prob.set_objective(self._v_wt[objective])
        return self._solve(lp.ObjectiveSense.Maximize)

    def get_fba_flux(self, objective):
        """Return a dictionary of all the fluxes solved by FBA.

        Dictionary of fluxes is used in :meth:`.lin_moma` and :meth:`.moma`
        to minimize changes in the flux distributions following model
        perturbation.

        Args:
            objective: The objective reaction that is maximized.

        Returns:
            Dictionary of fluxes for each reaction in the model.
        """
        flux_result = self.solve_fba(objective)
        fba_fluxes = {}

        # Place all the flux values in a dictionary
        for key in self._model.reactions:
            fba_fluxes[key] = flux_result.get_value(self._v_wt[key])
        return fba_fluxes

    def get_minimal_fba_flux(self, objective):
        """Find the FBA solution that minimizes all the flux values.

        Maximize the objective flux then minimize all other fluxes
        while keeping the objective flux at the maximum.

        Args:
            objective: The objective reaction that is maximized.

        Returns:
            A dictionary of all the reactions and their minimized fluxes.
        """
        # Define constraints
        vs_wt = self._v_wt.set(self._model.reactions)
        zs = self._z.set(self._model.reactions)

        wt_obj_flux = self.get_fba_obj_flux(objective)

        with self.constraints() as constr:
            constr.add(
                zs >= vs_wt, vs_wt >= -zs,
                self._v_wt[objective] >= wt_obj_flux)
            self._prob.set_objective(self._z.sum(self._model.reactions))
            result = self._solve(lp.ObjectiveSense.Minimize)

        fba_fluxes = {}
        for key in self._model.reactions:
            fba_fluxes[key] = result.get_value(self._v_wt[key])
        return fba_fluxes

    def get_fba_obj_flux(self, objective):
        """Return the maximum objective flux solved by FBA."""
        flux_result = self.solve_fba(objective)
        return flux_result.get_value(self._v_wt[objective])

    def lin_moma(self, wt_fluxes):
        """Minimize the redistribution of fluxes using a linear objective.

        The change in flux distribution is mimimized by minimizing the sum
        of the absolute values of the differences of wild type FBA solution
        and the knockout strain flux solution.

        This formulation bases the solution on the wild type fluxes that
        are specified by the user. If these wild type fluxes were calculated
        using FBA, then an arbitrary flux vector that optimizes the objective
        function is used. See [Segre`_02] for more information.

        Args:
            wt_fluxes: Dictionary of all the wild type fluxes. Use
                :meth:`.get_fba_flux(objective)` to return a dictionary of
                fluxes found by FBA.
        """
        reactions = set(self._adjustment_reactions())

        z_diff = self._z_diff
        v = self._v

        with self.constraints() as constr:
            for f_reaction, f_value in iteritems(wt_fluxes):
                if f_reaction in reactions:
                    # Add the constraint that finds the optimal solution, such
                    # that the difference between the wildtype flux is similar
                    # to the knockout flux.
                    constr.add(
                        z_diff[f_reaction] >= f_value - v[f_reaction],
                        f_value - v[f_reaction] >= -z_diff[f_reaction])

            # If we minimize the sum of the z vector then we will minimize
            # the |vs_wt - vs| from above
            self._prob.set_objective(z_diff.sum(reactions))

            self._solve(lp.ObjectiveSense.Minimize)

    def lin_moma2(self, objective, wt_obj):
        """Find the smallest redistribution vector using a linear objective.

        The change in flux distribution is mimimized by minimizing the sum
        of the absolute values of the differences of wild type FBA solution
        and the knockout strain flux solution.

        Creates the constraint that the we select the optimal flux vector that
        is closest to the wildtype. This might still return an arbitrary flux
        vector the maximizes the objective function.

        Args:
            objective: Objective reaction for the model.
            wt_obj: The flux value for your wild type objective reactions.
                Can either use an expiremental value or on determined by FBA
                by using :meth:`.get_fba_obj_flux(objective)`.
        """
        reactions = set(self._adjustment_reactions())

        z_diff = self._z_diff
        v = self._v
        v_wt = self._v_wt

        with self.constraints() as constr:
            for f_reaction in reactions:
                # Add the constraint that finds the optimal solution, such
                # that the difference between the wildtype flux
                # is similar to the knockout flux.
                constr.add(
                    z_diff[f_reaction] >= v_wt[f_reaction] - v[f_reaction],
                    v_wt[f_reaction] - v[f_reaction] >= -z_diff[f_reaction])

            # If we minimize the sum of the z vector then we will minimize
            # the |v_wt - v| from above
            self._prob.set_objective(z_diff.sum(reactions))

            constr.add(self._v_wt[objective] >= wt_obj)

            self._solve(lp.ObjectiveSense.Minimize)

    def moma(self, wt_fluxes):
        """Minimize the redistribution of fluxes using Euclidean distance.

        Minimizing the redistribution of fluxes using a quadratic objective
        function. The distance is minimized by minimizing the sum of
        (wild type - knockout)^2.

        Args:
            wt_fluxes: Dictionary of all the wild type fluxes that will be
                used to find a close MOMA solution. Fluxes can be expiremental
                or calculated using :meth: get_fba_flux(objective).
        """
        reactions = set(self._adjustment_reactions())
        v = self._v

        obj_expr = 0
        for f_reaction, f_value in iteritems(wt_fluxes):
            if f_reaction in reactions:
                # Minimize the Euclidean distance between the two vectors
                obj_expr += (f_value - v[f_reaction])**2

        self._prob.set_objective(obj_expr)
        self._solve(lp.ObjectiveSense.Minimize)

    def moma2(self, objective, wt_obj):
        """Find the smallest redistribution vector using Euclidean distance.

        Minimizing the redistribution of fluxes using a quadratic objective
        function. The distance is minimized by minimizing the sum of
        (wild type - knockout)^2.

        Creates the constraint that the we select the optimal flux vector that
        is closest to the wildtype. This might still return an arbitrary flux
        vector the maximizes the objective function.

        Args:
            objective: Objective reaction for the model.
            wt_obj: The flux value for your wild type objective reactions.
                Can either use an expiremental value or on determined by FBA
                by using :meth:`.get_fba_obj_flux(objective)`.
        """
        obj_expr = 0
        for reaction in self._adjustment_reactions():
            v_wt = self._v_wt[reaction]
            v = self._v[reaction]
            obj_expr += (v_wt - v)**2

        self._prob.set_objective(obj_expr)

        with self.constraints(self._v_wt[objective] >= wt_obj):
            self._solve(lp.ObjectiveSense.Minimize)

    def get_flux(self, reaction):
        """Return the knockout flux for a specific reaction."""
        return self._prob.result.get_value(self._v[reaction])

    def get_flux_var(self, reaction):
        """Return the LP variable for a specific reaction."""
        return self._v[reaction]



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2016  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Module related to loading ModelSEED database files."""

import csv
import re

from .context import FileMark
from .entry import CompoundEntry as BaseCompoundEntry


class ParseError(Exception):
    """Exception used to signal errors while parsing"""


def decode_name(s):
    """Decode names in ModelSEED files"""
    # Some names contain XML-like entity codes
    return re.sub(r'&#(\d+);', lambda x: chr(int(x.group(1))), s)


class CompoundEntry(BaseCompoundEntry):
    """Representation of entry in a ModelSEED compound table"""

    def __init__(self, id, names, formula, filemark=None):
        self._id = id
        self._properties = {
            'id': self._id,
            'names': list(names),
            'formula': formula
        }

        # Find shortest name
        # Usually the best name is the shortest one,
        # but in the case of ties, the best one is
        # usually the last one to appear in the list.
        # Since the sort is stable we can obtain this
        # by first reversing the list, then sorting
        # by length.
        names = self._properties['names']
        self._properties['name'] = None
        if len(self._properties['names']) > 0:
            name = sorted(reversed(names), key=lambda x: len(x))[0]
            self._properties['name'] = name

        self._filemark = filemark

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._properties.get('name')

    @property
    def names(self):
        return iter(self._properties.get('names'))

    @property
    def formula(self):
        return self._properties.get('formula')

    @property
    def properties(self):
        return dict(self._properties)

    @property
    def filemark(self):
        return self._filemark


def parse_compound_file(f, context=None):
    """Iterate over the compound entries in the given file"""

    f.readline()  # Skip header
    for lineno, row in enumerate(csv.reader(f, delimiter='\t')):
        compound_id, names, formula = row[:3]
        names = (decode_name(name) for name in names.split(',<br>'))

        # ModelSEED sometimes uses an asterisk and number at
        # the end of formulas. This seems to have a similar
        # meaning as '(...)n'.
        m = re.match(r'^(.*)\*(\d*)$', formula)
        if m is not None:
            if m.group(2) != '':
                formula = '({}){}'.format(m.group(1), m.group(2))
            else:
                formula = '({})n'.format(m.group(1))

        formula = formula.strip()
        if formula == '' or formula == 'noformula':
            formula = None

        mark = FileMark(context, lineno, 0)
        yield CompoundEntry(compound_id, names, formula, filemark=mark)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Linear programming solver using GLPK."""

from __future__ import absolute_import

import logging
from itertools import repeat, count
import numbers

from six.moves import zip

import swiglpk

from .lp import Solver as BaseSolver
from .lp import Constraint as BaseConstraint
from .lp import Problem as BaseProblem
from .lp import Result as BaseResult
from .lp import (Expression, RelationSense, ObjectiveSense, VariableType,
                 InvalidResultError, ranged_property, SolverError)

# Module-level logging
logger = logging.getLogger(__name__)

# Logger specific to log messages from GLPK library
_glpk_logger = logging.getLogger('glpk')


# Redirect GLPK terminal output to logger
def _term_hook(s):
    _glpk_logger.debug(s.rstrip())


swiglpk.glp_term_hook(_term_hook)


_INF = float('inf')


class GLPKError(SolverError):
    """Error from calling GLPK library."""


class Solver(BaseSolver):
    """Represents an LP-solver using GLPK."""

    def __init__(self):
        super(Solver, self).__init__()
        logger.warning('Support for GLPK solver is experimental!')

    def create_problem(self, **kwargs):
        """Create a new LP-problem using the solver."""
        return Problem(**kwargs)


class Problem(BaseProblem):
    """Represents an LP-problem of a :class:`.Solver`."""

    VARTYPE_MAP = {
        VariableType.Continuous: swiglpk.GLP_CV,
        VariableType.Binary: swiglpk.GLP_BV,
        VariableType.Integer: swiglpk.GLP_IV
    }

    OBJ_SENSE_MAP = {
        ObjectiveSense.Maximize: swiglpk.GLP_MAX,
        ObjectiveSense.Minimize: swiglpk.GLP_MIN
    }

    def __init__(self, **kwargs):
        self._p = swiglpk.glp_create_prob()

        self._variables = {}
        self._constraints = {}

        self._do_presolve = True

        # Initialize simplex tolerances to default values
        parm = swiglpk.glp_smcp()
        swiglpk.glp_init_smcp(parm)
        self._feasibility_tolerance = parm.tol_bnd
        self._optimality_tolerance = parm.tol_dj

        # Initialize mip tolerance to default value
        parm = swiglpk.glp_iocp()
        swiglpk.glp_init_iocp(parm)
        self._integrality_tolerance = parm.tol_int

        self._result = None

    def __del__(self):
        swiglpk.glp_delete_prob(self._p)

    @property
    def glpk(self):
        """The underlying GLPK (SWIG) object."""
        return self._p

    def define(self, *names, **kwargs):
        """Define a variable in the problem.

        Variables must be defined before they can be accessed by var() or
        set(). This function takes keyword arguments lower and upper to define
        the bounds of the variable (default: -inf to inf). The keyword argument
        types can be used to select the type of the variable (Continuous
        (default), Binary or Integer). Setting any variables different than
        Continuous will turn the problem into an MILP problem. Raises
        ValueError if a name is already defined.
        """
        names = tuple(names)
        for name in names:
            if name in self._variables:
                raise ValueError('Variable already defined: {!r}'.format(name))

        lower = kwargs.get('lower', None)
        upper = kwargs.get('upper', None)
        vartype = kwargs.get('types', None)

        # Repeat values if a scalar is given
        if lower is None or isinstance(lower, numbers.Number):
            lower = repeat(lower, len(names))
        if upper is None or isinstance(upper, numbers.Number):
            upper = repeat(upper, len(names))
        if vartype is None or vartype in (
                VariableType.Continuous, VariableType.Binary,
                VariableType.Integer):
            vartype = repeat(vartype, len(names))

        # Assign default values
        vartype = tuple(VariableType.Continuous if value is None else value
                        for value in vartype)

        if len(names) == 0:
            return

        var_indices = count(swiglpk.glp_add_cols(self._p, len(names)))

        for i, name, lb, ub, vt in zip(
                var_indices, names, lower, upper, vartype):
            self._variables[name] = i

            lb = None if lb == -_INF else lb
            ub = None if ub == _INF else ub

            if lb is None and ub is None:
                swiglpk.glp_set_col_bnds(self._p, i, swiglpk.GLP_FR, 0, 0)
            elif lb is None:
                swiglpk.glp_set_col_bnds(
                    self._p, i, swiglpk.GLP_UP, 0, float(ub))
            elif ub is None:
                swiglpk.glp_set_col_bnds(
                    self._p, i, swiglpk.GLP_LO, float(lb), 0)
            elif lb == ub:
                swiglpk.glp_set_col_bnds(
                    self._p, i, swiglpk.GLP_FX, float(lb), 0)
            else:
                swiglpk.glp_set_col_bnds(
                    self._p, i, swiglpk.GLP_DB, float(lb), float(ub))

            if vt != VariableType.Continuous:
                swiglpk.glp_set_col_kind(self._p, i, self.VARTYPE_MAP[vt])

        self._do_presolve = True

    def has_variable(self, name):
        """Check whether variable is defined in the model."""
        return name in self._variables

    def _add_constraints(self, relation):
        """Add the given relation as one or more constraints.

        Return a list of the names of the constraints added.
        """

        expression = relation.expression
        constr_count = sum(True for _ in expression.value_sets())
        if constr_count == 0:
            return []

        row_indices = count(swiglpk.glp_add_rows(self._p, constr_count))

        names = []
        for i, value_set in zip(row_indices, expression.value_sets()):
            value_set = list(value_set)
            var_indices = swiglpk.intArray(1 + len(value_set))
            var_values = swiglpk.doubleArray(1 + len(value_set))
            for j, (variable, coeff) in enumerate(value_set):
                var_indices[1 + j] = self._variables[variable]
                var_values[1 + j] = float(coeff)

            swiglpk.glp_set_mat_row(
                self._p, i, len(value_set), var_indices, var_values)

            if relation.sense == RelationSense.Greater:
                swiglpk.glp_set_row_bnds(
                    self._p, i, swiglpk.GLP_LO, -float(expression.offset), 0)
            elif relation.sense == RelationSense.Less:
                swiglpk.glp_set_row_bnds(
                    self._p, i, swiglpk.GLP_UP, 0, -float(expression.offset))
            else:
                swiglpk.glp_set_row_bnds(
                    self._p, i, swiglpk.GLP_FX, -float(expression.offset), 0)

            names.append(i)

        self._do_presolve = True

        return names

    def add_linear_constraints(self, *relations):
        """Add constraints to the problem.

        Each constraint is represented by a Relation, and the
        expression in that relation can be a set expression.
        """
        constraints = []

        for relation in relations:
            if self._check_relation(relation):
                constraints.append(Constraint(self, None))
            else:
                for name in self._add_constraints(relation):
                    constraints.append(Constraint(self, name))

        return constraints

    def set_objective(self, expression):
        """Set objective of problem."""

        if isinstance(expression, numbers.Number):
            # Allow expressions with no variables as objective,
            # represented as a number
            expression = Expression(offset=expression)

        # Clear previous objective
        for i in range(swiglpk.glp_get_num_cols(self._p)):
            swiglpk.glp_set_obj_coef(self._p, 1 + i, 0)

        for variable, value in expression.values():
            var_index = self._variables[variable]
            swiglpk.glp_set_obj_coef(self._p, var_index, float(value))

        swiglpk.glp_set_obj_coef(self._p, 0, float(expression.offset))

    set_linear_objective = set_objective
    """Set objective of the problem.

    .. deprecated:: 0.19
       Use :meth:`set_objective` instead.
    """

    def set_objective_sense(self, sense):
        """Set type of problem (maximize or minimize)."""

        if sense not in (ObjectiveSense.Minimize, ObjectiveSense.Maximize):
            raise ValueError('Invalid objective sense')

        swiglpk.glp_set_obj_dir(self._p, self.OBJ_SENSE_MAP[sense])

    def solve_unchecked(self, sense=None):
        """Solve problem and return result.

        The user must manually check the status of the result to determine
        whether an optimal solution was found. A :class:`SolverError` may still
        be raised if the underlying solver raises an exception.
        """
        if sense is not None:
            self.set_objective_sense(sense)

        parm = swiglpk.glp_smcp()
        swiglpk.glp_init_smcp(parm)
        if self._do_presolve:
            parm.presolve = swiglpk.GLP_ON
            self._do_presolve = False
        else:
            parm.presolve = swiglpk.GLP_OFF

        parm.tol_bnd = self._feasibility_tolerance
        parm.tol_dj = self._optimality_tolerance

        logger.debug('Solving using glp_simplex()')
        r = swiglpk.glp_simplex(self._p, parm)
        if r in (swiglpk.GLP_ENOPFS, swiglpk.GLP_ENODFS):
            self._result = Result(self, r)
            return self._result
        elif r != 0:
            raise GLPKError('glp_simplex: {}'.format(r))

        if swiglpk.glp_get_num_int(self._p) == 0:
            self._result = Result(self)
        else:
            # The intopt MILP solver needs an optimal solution to the LP
            # relaxation. Therefore, glp_simplex has to run before glp_intopt
            # for MILP problems.
            logger.debug('Solving using glp_intopt()')
            parm = swiglpk.glp_iocp()
            swiglpk.glp_init_iocp(parm)
            parm.tol_int = self._integrality_tolerance
            r = swiglpk.glp_intopt(self._p, parm)
            if r != 0:
                raise GLPKError('glp_intopt: {}'.format(r))

            self._result = MIPResult(self)

        return self._result

    @property
    def result(self):
        return self._result

    @ranged_property(min=None, max=None)
    def feasibility_tolerance(self):
        """Feasibility tolerance."""
        return self._feasibility_tolerance

    @feasibility_tolerance.setter
    def feasibility_tolerance(self, value):
        logger.info('Setting feasibility tolerance to {!r}'.format(value))
        self._feasibility_tolerance = value

    @ranged_property(min=None, max=None)
    def optimality_tolerance(self):
        """Optimality tolerance."""
        return self._optimality_tolerance

    @optimality_tolerance.setter
    def optimality_tolerance(self, value):
        logger.info('Setting optimality tolerance to {!r}'.format(value))
        self._optimality_tolerance = value

    @ranged_property(min=None, max=None)
    def integrality_tolerance(self):
        """Integrality tolerance."""
        return self._integrality_tolerance

    @integrality_tolerance.setter
    def integrality_tolerance(self, value):
        logger.info('Setting integrality tolerance to {!r}'.format(value))
        self._integrality_tolerance = value


class Constraint(BaseConstraint):
    """Represents a constraint in a :class:`.Problem`."""

    def __init__(self, prob, name):
        self._prob = prob
        self._name = name

    def delete(self):
        if self._name is not None:
            swiglpk.glp_set_row_bnds(
                self._prob._p, self._name, swiglpk.GLP_FR, 0, 0)


class Result(BaseResult):
    """Represents the solution to a :class:`.Problem`.

    This object will be returned from the solve() method on :class:`.Problem`
    or by accessing the result property on :class:`.Problem` after solving a
    problem. This class should not be instantiated manually.

    Result will evaluate to a boolean according to the success of the
    solution, so checking the truth value of the result will immediately
    indicate whether solving was successful.
    """

    def __init__(self, prob, ret_val=0):
        self._problem = prob
        self._ret_val = ret_val

    def _check_valid(self):
        if self._problem.result != self:
            raise InvalidResultError()

    @property
    def success(self):
        """Return boolean indicating whether a solution was found."""
        self._check_valid()
        if self._ret_val != 0:
            return False
        return swiglpk.glp_get_status(self._problem._p) == swiglpk.GLP_OPT

    @property
    def unbounded(self):
        """Whether solution is unbounded"""
        self._check_valid()
        if self._ret_val != 0:
            return self._ret_val == swiglpk.GLP_ENODFS
        return swiglpk.glp_get_status(self._problem._p) == swiglpk.GLP_UNBND

    @property
    def status(self):
        """Return string indicating the error encountered on failure."""
        self._check_valid()
        if self._ret_val == swiglpk.GLP_ENOPFS:
            return 'No primal feasible solution'
        elif self._ret_val == swiglpk.GLP_ENODFS:
            return 'No dual feasible solution'
        return str(swiglpk.glp_get_status(self._problem._p))

    def _has_variable(self, variable):
        """Whether variable exists in the solution."""
        return self._problem.has_variable(variable)

    def _get_value(self, variable):
        """Return value of variable in solution."""
        return swiglpk.glp_get_col_prim(
            self._problem._p, self._problem._variables[variable])

    def get_value(self, expression):
        """Return value of expression."""
        self._check_valid()
        return super(Result, self).get_value(expression)


class MIPResult(Result):
    """Specialization of Result for MIP problems."""

    @property
    def success(self):
        self._check_valid()
        return swiglpk.glp_mip_status(self._problem._p) == swiglpk.GLP_OPT

    @property
    def status(self):
        self._check_valid()
        return str(swiglpk.glp_mip_status(self._problem._p))

    def _get_value(self, variable):
        """Return value of variable in solution."""
        return swiglpk.glp_mip_col_val(
            self._problem._p, self._problem._variables[variable])
